# Copyright 2018 Huawei
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import http.client
import uuid

from keystone.common import provider_api
from keystone.common.validation import validators
import keystone.conf
from keystone.tests import unit
from keystone.tests.unit import test_v3

CONF = keystone.conf.CONF
PROVIDERS = provider_api.ProviderAPIs


class LimitModelTestCase(test_v3.RestfulTestCase):

    def test_get_default_limit_model_response_schema(self):
        schema = {
            'type': 'object',
            'properties': {
                'model': {
                    'type': 'object',
                    'properties': {
                        'name': {'type': 'string'},
                        'description': {'type': 'string'}
                    },
                    'required': ['name', 'description'],
                    'additionalProperties': False,
                },
            },
            'required': ['model'],
            'additionalProperties': False,
        }
        validator = validators.SchemaValidator(schema)
        response = self.get('/limits/model')
        validator.validate(response.json_body)

    def test_head_limit_model(self):
        self.head('/limits/model', expected_status=http.client.OK)

    def test_get_limit_model_returns_default_model(self):
        response = self.get('/limits/model')
        model = response.result
        expected = {
            'model': {
                'name': 'flat',
                'description': (
                    'Limit enforcement and validation does not take project '
                    'hierarchy into consideration.'
                )
            }
        }
        self.assertDictEqual(expected, model)

    def test_get_limit_model_without_token_fails(self):
        self.get(
            '/limits/model', noauth=True,
            expected_status=http.client.UNAUTHORIZED
        )

    def test_head_limit_model_without_token_fails(self):
        self.head(
            '/limits/model', noauth=True,
            expected_status=http.client.UNAUTHORIZED
        )


class RegisteredLimitsTestCase(test_v3.RestfulTestCase):
    """Test registered_limits CRUD."""

    def setUp(self):
        super(RegisteredLimitsTestCase, self).setUp()

        # Most of these tests require system-scoped tokens. Let's have one on
        # hand so that we can use it in tests when we need it.
        PROVIDERS.assignment_api.create_system_grant_for_user(
            self.user_id, self.role_id
        )
        self.system_admin_token = self.get_system_scoped_token()

        # There is already a sample service and region created from
        # load_sample_data() but we're going to create another service and
        # region for specific testing purposes.
        response = self.post('/regions', body={'region': {}})
        self.region2 = response.json_body['region']
        self.region_id2 = self.region2['id']

        service_ref = {'service': {
            'name': uuid.uuid4().hex,
            'enabled': True,
            'type': 'type2'}}
        response = self.post('/services', body=service_ref)
        self.service2 = response.json_body['service']
        self.service_id2 = self.service2['id']

    def test_create_registered_limit(self):
        ref = unit.new_registered_limit_ref(service_id=self.service_id,
                                            region_id=self.region_id)
        r = self.post(
            '/registered_limits',
            body={'registered_limits': [ref]},
            token=self.system_admin_token,
            expected_status=http.client.CREATED)
        registered_limits = r.result['registered_limits']
        for key in ['service_id', 'region_id', 'resource_name',
                    'default_limit', 'description']:
            self.assertEqual(registered_limits[0][key], ref[key])

    def test_create_registered_limit_without_region(self):
        ref = unit.new_registered_limit_ref(service_id=self.service_id)
        r = self.post(
            '/registered_limits',
            body={'registered_limits': [ref]},
            token=self.system_admin_token,
            expected_status=http.client.CREATED)
        registered_limits = r.result['registered_limits']
        for key in ['service_id', 'resource_name', 'default_limit']:
            self.assertEqual(registered_limits[0][key], ref[key])
        self.assertIsNone(registered_limits[0].get('region_id'))

    def test_create_registered_without_description(self):
        ref = unit.new_registered_limit_ref(service_id=self.service_id,
                                            region_id=self.region_id)
        ref.pop('description')
        r = self.post(
            '/registered_limits',
            body={'registered_limits': [ref]},
            token=self.system_admin_token,
            expected_status=http.client.CREATED)
        registered_limits = r.result['registered_limits']
        for key in ['service_id', 'region_id', 'resource_name',
                    'default_limit']:
            self.assertEqual(registered_limits[0][key], ref[key])
        self.assertIsNone(registered_limits[0]['description'])

    def test_create_multi_registered_limit(self):
        ref1 = unit.new_registered_limit_ref(service_id=self.service_id,
                                             region_id=self.region_id,
                                             resource_name='volume')
        ref2 = unit.new_registered_limit_ref(service_id=self.service_id,
                                             resource_name='snapshot')
        r = self.post(
            '/registered_limits',
            body={'registered_limits': [ref1, ref2]},
            token=self.system_admin_token,
            expected_status=http.client.CREATED)
        registered_limits = r.result['registered_limits']
        for key in ['service_id', 'resource_name', 'default_limit']:
            self.assertEqual(registered_limits[0][key], ref1[key])
            self.assertEqual(registered_limits[1][key], ref2[key])
        self.assertEqual(registered_limits[0]['region_id'], ref1['region_id'])
        self.assertIsNone(registered_limits[1].get('region_id'))

    def test_create_registered_limit_return_count(self):
        ref1 = unit.new_registered_limit_ref(service_id=self.service_id,
                                             region_id=self.region_id)
        r = self.post(
            '/registered_limits',
            body={'registered_limits': [ref1]},
            token=self.system_admin_token,
            expected_status=http.client.CREATED)
        registered_limits = r.result['registered_limits']
        self.assertEqual(1, len(registered_limits))

        ref2 = unit.new_registered_limit_ref(service_id=self.service_id2,
                                             region_id=self.region_id2)
        ref3 = unit.new_registered_limit_ref(service_id=self.service_id2)
        r = self.post(
            '/registered_limits',
            body={'registered_limits': [ref2, ref3]},
            token=self.system_admin_token,
            expected_status=http.client.CREATED)
        registered_limits = r.result['registered_limits']
        self.assertEqual(2, len(registered_limits))

    def test_create_registered_limit_with_invalid_input(self):
        ref1 = unit.new_registered_limit_ref()
        ref2 = unit.new_registered_limit_ref(default_limit='not_int')
        ref3 = unit.new_registered_limit_ref(resource_name=123)
        ref4 = unit.new_registered_limit_ref(region_id='fake_region')
        for input_limit in [ref1, ref2, ref3, ref4]:
            self.post(
                '/registered_limits',
                body={'registered_limits': [input_limit]},
                token=self.system_admin_token,
                expected_status=http.client.BAD_REQUEST)

    def test_create_registered_limit_duplicate(self):
        ref = unit.new_registered_limit_ref(service_id=self.service_id,
                                            region_id=self.region_id)
        self.post(
            '/registered_limits',
            body={'registered_limits': [ref]},
            token=self.system_admin_token,
            expected_status=http.client.CREATED)
        self.post(
            '/registered_limits',
            body={'registered_limits': [ref]},
            token=self.system_admin_token,
            expected_status=http.client.CONFLICT)

    def test_update_registered_limit(self):
        ref = unit.new_registered_limit_ref(service_id=self.service_id,
                                            region_id=self.region_id,
                                            resource_name='volume',
                                            default_limit=10)
        r = self.post(
            '/registered_limits',
            body={'registered_limits': [ref]},
            token=self.system_admin_token,
            expected_status=http.client.CREATED)
        update_ref = {
            'service_id': self.service_id2,
            'region_id': self.region_id2,
            'resource_name': 'snapshot',
            'default_limit': 5,
            'description': 'test description'
        }
        r = self.patch(
            '/registered_limits/%s' % r.result['registered_limits'][0]['id'],
            body={'registered_limit': update_ref},
            token=self.system_admin_token,
            expected_status=http.client.OK)
        new_registered_limits = r.result['registered_limit']

        self.assertEqual(new_registered_limits['service_id'], self.service_id2)
        self.assertEqual(new_registered_limits['region_id'], self.region_id2)
        self.assertEqual(new_registered_limits['resource_name'], 'snapshot')
        self.assertEqual(new_registered_limits['default_limit'], 5)
        self.assertEqual(new_registered_limits['description'],
                         'test description')

    def test_update_registered_limit_region_failed(self):
        ref = unit.new_registered_limit_ref(service_id=self.service_id,
                                            resource_name='volume',
                                            default_limit=10,
                                            description='test description')
        r = self.post(
            '/registered_limits',
            body={'registered_limits': [ref]},
            token=self.system_admin_token,
            expected_status=http.client.CREATED)
        update_ref = {
            'region_id': self.region_id,
        }
        registered_limit_id = r.result['registered_limits'][0]['id']
        r = self.patch(
            '/registered_limits/%s' % registered_limit_id,
            body={'registered_limit': update_ref},
            token=self.system_admin_token,
            expected_status=http.client.OK)
        new_registered_limits = r.result['registered_limit']
        self.assertEqual(self.region_id, new_registered_limits['region_id'])

        update_ref['region_id'] = ''
        r = self.patch(
            '/registered_limits/%s' % registered_limit_id,
            body={'registered_limit': update_ref},
            token=self.system_admin_token,
            expected_status=http.client.BAD_REQUEST)

    def test_update_registered_limit_description(self):
        ref = unit.new_registered_limit_ref(service_id=self.service_id,
                                            region_id=self.region_id,
                                            resource_name='volume',
                                            default_limit=10)
        r = self.post(
            '/registered_limits',
            body={'registered_limits': [ref]},
            token=self.system_admin_token,
            expected_status=http.client.CREATED)
        update_ref = {
            'description': 'test description'
        }
        registered_limit_id = r.result['registered_limits'][0]['id']
        r = self.patch(
            '/registered_limits/%s' % registered_limit_id,
            body={'registered_limit': update_ref},
            token=self.system_admin_token,
            expected_status=http.client.OK)
        new_registered_limits = r.result['registered_limit']
        self.assertEqual(new_registered_limits['description'],
                         'test description')

        update_ref['description'] = ''
        r = self.patch(
            '/registered_limits/%s' % registered_limit_id,
            body={'registered_limit': update_ref},
            token=self.system_admin_token,
            expected_status=http.client.OK)
        new_registered_limits = r.result['registered_limit']
        self.assertEqual(new_registered_limits['description'], '')

    def test_update_registered_limit_region_id_to_none(self):
        ref = unit.new_registered_limit_ref(service_id=self.service_id,
                                            region_id=self.region_id,
                                            resource_name='volume',
                                            default_limit=10)
        r = self.post(
            '/registered_limits',
            body={'registered_limits': [ref]},
            token=self.system_admin_token,
            expected_status=http.client.CREATED)
        update_ref = {
            'region_id': None
        }
        registered_limit_id = r.result['registered_limits'][0]['id']
        r = self.patch(
            '/registered_limits/%s' % registered_limit_id,
            body={'registered_limit': update_ref},
            token=self.system_admin_token,
            expected_status=http.client.OK)
        self.assertIsNone(r.result['registered_limit']['region_id'])

    def test_update_registered_limit_region_id_to_none_conflict(self):
        ref1 = unit.new_registered_limit_ref(service_id=self.service_id,
                                             resource_name='volume',
                                             default_limit=10)
        ref2 = unit.new_registered_limit_ref(service_id=self.service_id,
                                             region_id=self.region_id,
                                             resource_name='volume',
                                             default_limit=10)
        self.post(
            '/registered_limits',
            body={'registered_limits': [ref1]},
            token=self.system_admin_token,
            expected_status=http.client.CREATED)
        r = self.post(
            '/registered_limits',
            body={'registered_limits': [ref2]},
            token=self.system_admin_token,
            expected_status=http.client.CREATED)

        update_ref = {
            'region_id': None
        }
        registered_limit_id = r.result['registered_limits'][0]['id']
        # There is a registered limit with "service_id=self.service_id,
        # region_id=None" already. So update ref2's region_id to None will
        # raise 409 Conflict Error.
        self.patch(
            '/registered_limits/%s' % registered_limit_id,
            body={'registered_limit': update_ref},
            token=self.system_admin_token,
            expected_status=http.client.CONFLICT)

    def test_update_registered_limit_not_found(self):
        update_ref = {
            'service_id': self.service_id,
            'region_id': self.region_id,
            'resource_name': 'snapshot',
            'default_limit': 5
        }
        self.patch(
            '/registered_limits/%s' % uuid.uuid4().hex,
            body={'registered_limit': update_ref},
            token=self.system_admin_token,
            expected_status=http.client.NOT_FOUND)

    def test_update_registered_limit_with_invalid_input(self):
        ref = unit.new_registered_limit_ref(service_id=self.service_id,
                                            region_id=self.region_id,
                                            resource_name='volume',
                                            default_limit=10)
        r = self.post(
            '/registered_limits',
            body={'registered_limits': [ref]},
            token=self.system_admin_token,
            expected_status=http.client.CREATED)
        reg_id = r.result['registered_limits'][0]['id']

        update_ref1 = unit.new_registered_limit_ref(service_id='fake_id')
        update_ref2 = unit.new_registered_limit_ref(default_limit='not_int')
        update_ref3 = unit.new_registered_limit_ref(resource_name=123)
        update_ref4 = unit.new_registered_limit_ref(region_id='fake_region')
        update_ref5 = unit.new_registered_limit_ref(description=123)
        for input_limit in [update_ref1, update_ref2, update_ref3,
                            update_ref4, update_ref5]:
            self.patch(
                '/registered_limits/%s' % reg_id,
                body={'registered_limit': input_limit},
                token=self.system_admin_token,
                expected_status=http.client.BAD_REQUEST)

    def test_update_registered_limit_with_referenced_limit(self):
        ref = unit.new_registered_limit_ref(service_id=self.service_id,
                                            region_id=self.region_id,
                                            resource_name='volume',
                                            default_limit=10)
        r = self.post(
            '/registered_limits',
            body={'registered_limits': [ref]},
            token=self.system_admin_token,
            expected_status=http.client.CREATED)

        ref = unit.new_limit_ref(project_id=self.project_id,
                                 service_id=self.service_id,
                                 region_id=self.region_id,
                                 resource_name='volume')
        self.post(
            '/limits',
            body={'limits': [ref]},
            token=self.system_admin_token,
            expected_status=http.client.CREATED)

        update_ref = {
            'service_id': self.service_id2,
            'region_id': self.region_id2,
            'resource_name': 'snapshot',
            'default_limit': 5
        }
        self.patch(
            '/registered_limits/%s' % r.result['registered_limits'][0]['id'],
            body={'registered_limit': update_ref},
            token=self.system_admin_token,
            expected_status=http.client.FORBIDDEN)

    def test_list_registered_limit(self):
        r = self.get(
            '/registered_limits',
            expected_status=http.client.OK)
        self.assertEqual([], r.result.get('registered_limits'))

        ref1 = unit.new_registered_limit_ref(service_id=self.service_id,
                                             resource_name='test_resource',
                                             region_id=self.region_id)
        ref2 = unit.new_registered_limit_ref(service_id=self.service_id2,
                                             resource_name='test_resource',
                                             region_id=self.region_id2)
        r = self.post(
            '/registered_limits',
            body={'registered_limits': [ref1, ref2]},
            token=self.system_admin_token,
            expected_status=http.client.CREATED)
        id1 = r.result['registered_limits'][0]['id']
        r = self.get(
            '/registered_limits',
            expected_status=http.client.OK)
        registered_limits = r.result['registered_limits']
        self.assertEqual(len(registered_limits), 2)
        for key in ['service_id', 'region_id', 'resource_name',
                    'default_limit']:
            if registered_limits[0]['id'] == id1:
                self.assertEqual(registered_limits[0][key], ref1[key])
                self.assertEqual(registered_limits[1][key], ref2[key])
                break
            self.assertEqual(registered_limits[1][key], ref1[key])
            self.assertEqual(registered_limits[0][key], ref2[key])

        r = self.get(
            '/registered_limits?service_id=%s' % self.service_id,
            expected_status=http.client.OK)
        registered_limits = r.result['registered_limits']
        self.assertEqual(len(registered_limits), 1)
        for key in ['service_id', 'region_id', 'resource_name',
                    'default_limit']:
            self.assertEqual(registered_limits[0][key], ref1[key])

        r = self.get(
            '/registered_limits?region_id=%s' % self.region_id2,
            expected_status=http.client.OK)
        registered_limits = r.result['registered_limits']
        self.assertEqual(len(registered_limits), 1)
        for key in ['service_id', 'region_id', 'resource_name',
                    'default_limit']:
            self.assertEqual(registered_limits[0][key], ref2[key])

        r = self.get(
            '/registered_limits?resource_name=test_resource',
            expected_status=http.client.OK)
        registered_limits = r.result['registered_limits']
        self.assertEqual(len(registered_limits), 2)

    def test_show_registered_limit(self):
        ref1 = unit.new_registered_limit_ref(service_id=self.service_id,
                                             region_id=self.region_id)
        ref2 = unit.new_registered_limit_ref(service_id=self.service_id2,
                                             region_id=self.region_id2)
        r = self.post(
            '/registered_limits',
            body={'registered_limits': [ref1, ref2]},
            token=self.system_admin_token,
            expected_status=http.client.CREATED)
        id1 = r.result['registered_limits'][0]['id']
        self.get(
            '/registered_limits/fake_id',
            expected_status=http.client.NOT_FOUND)
        r = self.get(
            '/registered_limits/%s' % id1,
            expected_status=http.client.OK)
        registered_limit = r.result['registered_limit']
        for key in ['service_id', 'region_id', 'resource_name',
                    'default_limit', 'description']:
            self.assertEqual(registered_limit[key], ref1[key])

    def test_delete_registered_limit(self):
        ref1 = unit.new_registered_limit_ref(service_id=self.service_id,
                                             region_id=self.region_id)
        ref2 = unit.new_registered_limit_ref(service_id=self.service_id2,
                                             region_id=self.region_id2)
        r = self.post(
            '/registered_limits',
            body={'registered_limits': [ref1, ref2]},
            token=self.system_admin_token,
            expected_status=http.client.CREATED)
        id1 = r.result['registered_limits'][0]['id']
        self.delete('/registered_limits/%s' % id1,
                    token=self.system_admin_token,
                    expected_status=http.client.NO_CONTENT)
        self.delete('/registered_limits/fake_id',
                    token=self.system_admin_token,
                    expected_status=http.client.NOT_FOUND)
        r = self.get(
            '/registered_limits',
            expected_status=http.client.OK)
        registered_limits = r.result['registered_limits']
        self.assertEqual(len(registered_limits), 1)

    def test_delete_registered_limit_with_referenced_limit(self):
        ref = unit.new_registered_limit_ref(service_id=self.service_id,
                                            region_id=self.region_id,
                                            resource_name='volume',
                                            default_limit=10)
        r = self.post(
            '/registered_limits',
            body={'registered_limits': [ref]},
            token=self.system_admin_token,
            expected_status=http.client.CREATED)

        ref = unit.new_limit_ref(project_id=self.project_id,
                                 service_id=self.service_id,
                                 region_id=self.region_id,
                                 resource_name='volume')
        self.post(
            '/limits',
            body={'limits': [ref]},
            token=self.system_admin_token,
            expected_status=http.client.CREATED)

        id = r.result['registered_limits'][0]['id']
        self.delete('/registered_limits/%s' % id,
                    expected_status=http.client.FORBIDDEN)


class LimitsTestCase(test_v3.RestfulTestCase):
    """Test limits CRUD."""

    def setUp(self):
        super(LimitsTestCase, self).setUp()
        # FIXME(lbragstad): Remove all this duplicated logic once we get all
        # keystone tests using bootstrap consistently. This is something the
        # bootstrap utility already does for us.
        reader_role = {'id': uuid.uuid4().hex, 'name': 'reader'}
        reader_role = PROVIDERS.role_api.create_role(
            reader_role['id'], reader_role
        )

        member_role = {'id': uuid.uuid4().hex, 'name': 'member'}
        member_role = PROVIDERS.role_api.create_role(
            member_role['id'], member_role
        )
        PROVIDERS.role_api.create_implied_role(self.role_id, member_role['id'])
        PROVIDERS.role_api.create_implied_role(
            member_role['id'], reader_role['id']
        )

        # Most of these tests require system-scoped tokens. Let's have one on
        # hand so that we can use it in tests when we need it.
        PROVIDERS.assignment_api.create_system_grant_for_user(
            self.user_id, self.role_id
        )
        self.system_admin_token = self.get_system_scoped_token()

        # There is already a sample service and region created from
        # load_sample_data() but we're going to create another service and
        # region for specific testing purposes.
        response = self.post('/regions', body={'region': {}})
        self.region2 = response.json_body['region']
        self.region_id2 = self.region2['id']

        service_ref = {'service': {
            'name': uuid.uuid4().hex,
            'enabled': True,
            'type': 'type2'}}
        response = self.post('/services', body=service_ref)
        self.service2 = response.json_body['service']
        self.service_id2 = self.service2['id']

        ref1 = unit.new_registered_limit_ref(service_id=self.service_id,
                                             region_id=self.region_id,
                                             resource_name='volume')
        ref2 = unit.new_registered_limit_ref(service_id=self.service_id2,
                                             resource_name='snapshot')
        ref3 = unit.new_registered_limit_ref(service_id=self.service_id,
                                             region_id=self.region_id,
                                             resource_name='backup')
        self.post(
            '/registered_limits',
            body={'registered_limits': [ref1, ref2, ref3]},
            token=self.system_admin_token,
            expected_status=http.client.CREATED)

        # Create more assignments, all are:
        #
        # self.user -- admin -- self.project
        # self.user -- non-admin -- self.project_2
        # self.user -- admin -- self.domain
        # self.user -- non-admin -- self.domain_2
        # self.user -- admin -- system
        self.project_2 = unit.new_project_ref(domain_id=self.domain_id)
        self.project_2_id = self.project_2['id']
        PROVIDERS.resource_api.create_project(self.project_2_id,
                                              self.project_2)

        self.domain_2 = unit.new_domain_ref()
        self.domain_2_id = self.domain_2['id']
        PROVIDERS.resource_api.create_domain(self.domain_2_id, self.domain_2)

        self.role_2 = unit.new_role_ref(name='non-admin')
        self.role_2_id = self.role_2['id']
        PROVIDERS.role_api.create_role(self.role_2_id, self.role_2)

        PROVIDERS.assignment_api.create_grant(
            self.role_2_id, user_id=self.user_id, project_id=self.project_2_id)
        PROVIDERS.assignment_api.create_grant(
            self.role_id, user_id=self.user_id, domain_id=self.domain_id)
        PROVIDERS.assignment_api.create_grant(
            self.role_2_id, user_id=self.user_id, domain_id=self.domain_2_id)
        PROVIDERS.assignment_api.create_system_grant_for_user(
            self.user_id, self.role_id)

    def test_create_project_limit(self):
        ref = unit.new_limit_ref(project_id=self.project_id,
                                 service_id=self.service_id,
                                 region_id=self.region_id,
                                 resource_name='volume')
        r = self.post(
            '/limits',
            body={'limits': [ref]},
            token=self.system_admin_token,
            expected_status=http.client.CREATED)
        limits = r.result['limits']

        self.assertIsNotNone(limits[0]['id'])
        self.assertIsNone(limits[0]['domain_id'])
        for key in ['service_id', 'region_id', 'resource_name',
                    'resource_limit', 'description', 'project_id']:
            self.assertEqual(limits[0][key], ref[key])

    def test_create_domain_limit(self):
        ref = unit.new_limit_ref(domain_id=self.domain_id,
                                 service_id=self.service_id,
                                 region_id=self.region_id,
                                 resource_name='volume')
        r = self.post(
            '/limits',
            body={'limits': [ref]},
            token=self.system_admin_token,
            expected_status=http.client.CREATED)
        limits = r.result['limits']

        self.assertIsNotNone(limits[0]['id'])
        self.assertIsNone(limits[0]['project_id'])
        for key in ['service_id', 'region_id', 'resource_name',
                    'resource_limit', 'description', 'domain_id']:
            self.assertEqual(limits[0][key], ref[key])

    def test_create_limit_without_region(self):
        ref = unit.new_limit_ref(project_id=self.project_id,
                                 service_id=self.service_id2,
                                 resource_name='snapshot')
        r = self.post(
            '/limits',
            body={'limits': [ref]},
            token=self.system_admin_token,
            expected_status=http.client.CREATED)
        limits = r.result['limits']

        self.assertIsNotNone(limits[0]['id'])
        self.assertIsNotNone(limits[0]['project_id'])
        for key in ['service_id', 'resource_name', 'resource_limit']:
            self.assertEqual(limits[0][key], ref[key])
        self.assertIsNone(limits[0].get('region_id'))

    def test_create_limit_without_description(self):
        ref = unit.new_limit_ref(project_id=self.project_id,
                                 service_id=self.service_id,
                                 region_id=self.region_id,
                                 resource_name='volume')
        ref.pop('description')
        r = self.post(
            '/limits',
            body={'limits': [ref]},
            token=self.system_admin_token,
            expected_status=http.client.CREATED)
        limits = r.result['limits']

        self.assertIsNotNone(limits[0]['id'])
        self.assertIsNotNone(limits[0]['project_id'])
        for key in ['service_id', 'region_id', 'resource_name',
                    'resource_limit']:
            self.assertEqual(limits[0][key], ref[key])
        self.assertIsNone(limits[0]['description'])

    def test_create_limit_with_domain_as_project(self):
        ref = unit.new_limit_ref(project_id=self.domain_id,
                                 service_id=self.service_id,
                                 region_id=self.region_id,
                                 resource_name='volume')
        r = self.post('/limits', body={'limits': [ref]},
                      token=self.system_admin_token)
        limits = r.result['limits']
        self.assertIsNone(limits[0]['project_id'])
        self.assertEqual(self.domain_id, limits[0]['domain_id'])

    def test_create_multi_limit(self):
        ref1 = unit.new_limit_ref(project_id=self.project_id,
                                  service_id=self.service_id,
                                  region_id=self.region_id,
                                  resource_name='volume')
        ref2 = unit.new_limit_ref(project_id=self.project_id,
                                  service_id=self.service_id2,
                                  resource_name='snapshot')
        r = self.post(
            '/limits',
            body={'limits': [ref1, ref2]},
            token=self.system_admin_token,
            expected_status=http.client.CREATED)
        limits = r.result['limits']
        for key in ['service_id', 'resource_name', 'resource_limit']:
            self.assertEqual(limits[0][key], ref1[key])
            self.assertEqual(limits[1][key], ref2[key])
        self.assertEqual(limits[0]['region_id'], ref1['region_id'])
        self.assertIsNone(limits[1].get('region_id'))

    def test_create_limit_return_count(self):
        ref1 = unit.new_limit_ref(project_id=self.project_id,
                                  service_id=self.service_id,
                                  region_id=self.region_id,
                                  resource_name='volume')
        r = self.post(
            '/limits',
            body={'limits': [ref1]},
            token=self.system_admin_token,
            expected_status=http.client.CREATED)
        limits = r.result['limits']
        self.assertEqual(1, len(limits))

        ref2 = unit.new_limit_ref(project_id=self.project_id,
                                  service_id=self.service_id2,
                                  resource_name='snapshot')
        ref3 = unit.new_limit_ref(project_id=self.project_id,
                                  service_id=self.service_id,
                                  region_id=self.region_id,
                                  resource_name='backup')
        r = self.post(
            '/limits',
            body={'limits': [ref2, ref3]},
            token=self.system_admin_token,
            expected_status=http.client.CREATED)
        limits = r.result['limits']
        self.assertEqual(2, len(limits))

    def test_create_limit_with_invalid_input(self):
        ref1 = unit.new_limit_ref(project_id=self.project_id,
                                  resource_limit='not_int')
        ref2 = unit.new_limit_ref(project_id=self.project_id,
                                  resource_name=123)
        ref3 = unit.new_limit_ref(project_id=self.project_id,
                                  region_id='fake_region')
        for input_limit in [ref1, ref2, ref3]:
            self.post(
                '/limits',
                body={'limits': [input_limit]},
                token=self.system_admin_token,
                expected_status=http.client.BAD_REQUEST)

    def test_create_limit_duplicate(self):
        ref = unit.new_limit_ref(project_id=self.project_id,
                                 service_id=self.service_id,
                                 region_id=self.region_id,
                                 resource_name='volume')
        self.post(
            '/limits',
            body={'limits': [ref]},
            token=self.system_admin_token,
            expected_status=http.client.CREATED)
        self.post(
            '/limits',
            body={'limits': [ref]},
            token=self.system_admin_token,
            expected_status=http.client.CONFLICT)

    def test_create_limit_without_reference_registered_limit(self):
        ref = unit.new_limit_ref(project_id=self.project_id,
                                 service_id=self.service_id,
                                 region_id=self.region_id2,
                                 resource_name='volume')
        self.post(
            '/limits',
            body={'limits': [ref]},
            token=self.system_admin_token,
            expected_status=http.client.FORBIDDEN)

    def test_update_limit(self):
        ref = unit.new_limit_ref(project_id=self.project_id,
                                 service_id=self.service_id,
                                 region_id=self.region_id,
                                 resource_name='volume',
                                 resource_limit=10)
        r = self.post(
            '/limits',
            body={'limits': [ref]},
            token=self.system_admin_token,
            expected_status=http.client.CREATED)
        update_ref = {
            'resource_limit': 5,
            'description': 'test description'
        }
        r = self.patch(
            '/limits/%s' % r.result['limits'][0]['id'],
            body={'limit': update_ref},
            token=self.system_admin_token,
            expected_status=http.client.OK)
        new_limits = r.result['limit']

        self.assertEqual(new_limits['resource_limit'], 5)
        self.assertEqual(new_limits['description'], 'test description')

    def test_update_limit_not_found(self):
        update_ref = {
            'resource_limit': 5
        }
        self.patch(
            '/limits/%s' % uuid.uuid4().hex,
            body={'limit': update_ref},
            token=self.system_admin_token,
            expected_status=http.client.NOT_FOUND)

    def test_update_limit_with_invalid_input(self):
        ref = unit.new_limit_ref(project_id=self.project_id,
                                 service_id=self.service_id,
                                 region_id=self.region_id,
                                 resource_name='volume',
                                 resource_limit=10)
        r = self.post(
            '/limits',
            body={'limits': [ref]},
            token=self.system_admin_token,
            expected_status=http.client.CREATED)
        limit_id = r.result['limits'][0]['id']

        invalid_resource_limit_update = {
            'resource_limit': 'not_int'
        }
        invalid_description_update = {
            'description': 123
        }
        for input_limit in [invalid_resource_limit_update,
                            invalid_description_update]:
            self.patch(
                '/limits/%s' % limit_id,
                body={'limit': input_limit},
                token=self.system_admin_token,
                expected_status=http.client.BAD_REQUEST)

    def test_list_limit(self):
        r = self.get(
            '/limits',
            token=self.system_admin_token,
            expected_status=http.client.OK)
        self.assertEqual([], r.result.get('limits'))

        ref1 = unit.new_limit_ref(project_id=self.project_id,
                                  service_id=self.service_id,
                                  region_id=self.region_id,
                                  resource_name='volume')
        ref2 = unit.new_limit_ref(project_id=self.project_id,
                                  service_id=self.service_id2,
                                  resource_name='snapshot')
        r = self.post(
            '/limits',
            body={'limits': [ref1, ref2]},
            token=self.system_admin_token,
            expected_status=http.client.CREATED)
        id1 = r.result['limits'][0]['id']
        r = self.get(
            '/limits',
            expected_status=http.client.OK)
        limits = r.result['limits']
        self.assertEqual(len(limits), 2)
        if limits[0]['id'] == id1:
            self.assertEqual(limits[0]['region_id'], ref1['region_id'])
            self.assertIsNone(limits[1].get('region_id'))
            for key in ['service_id', 'resource_name', 'resource_limit']:
                self.assertEqual(limits[0][key], ref1[key])
                self.assertEqual(limits[1][key], ref2[key])
        else:
            self.assertEqual(limits[1]['region_id'], ref1['region_id'])
            self.assertIsNone(limits[0].get('region_id'))
            for key in ['service_id', 'resource_name', 'resource_limit']:
                self.assertEqual(limits[1][key], ref1[key])
                self.assertEqual(limits[0][key], ref2[key])

        r = self.get(
            '/limits?service_id=%s' % self.service_id2,
            expected_status=http.client.OK)
        limits = r.result['limits']
        self.assertEqual(len(limits), 1)
        for key in ['service_id', 'resource_name', 'resource_limit']:
            self.assertEqual(limits[0][key], ref2[key])

        r = self.get(
            '/limits?region_id=%s' % self.region_id,
            expected_status=http.client.OK)
        limits = r.result['limits']
        self.assertEqual(len(limits), 1)
        for key in ['service_id', 'region_id', 'resource_name',
                    'resource_limit']:
            self.assertEqual(limits[0][key], ref1[key])

        r = self.get(
            '/limits?resource_name=volume',
            expected_status=http.client.OK)
        limits = r.result['limits']
        self.assertEqual(len(limits), 1)
        for key in ['service_id', 'region_id', 'resource_name',
                    'resource_limit']:
            self.assertEqual(limits[0][key], ref1[key])

    def test_list_limit_with_project_id_filter(self):
        # create two limit in different projects for test.
        self.config_fixture.config(group='oslo_policy',
                                   enforce_scope=True)
        ref1 = unit.new_limit_ref(project_id=self.project_id,
                                  service_id=self.service_id,
                                  region_id=self.region_id,
                                  resource_name='volume')
        ref2 = unit.new_limit_ref(project_id=self.project_2_id,
                                  service_id=self.service_id2,
                                  resource_name='snapshot')
        self.post(
            '/limits',
            body={'limits': [ref1, ref2]},
            token=self.system_admin_token,
            expected_status=http.client.CREATED)

        # non system scoped request will get the limits in its project.
        r = self.get('/limits', expected_status=http.client.OK)
        limits = r.result['limits']
        self.assertEqual(1, len(limits))
        self.assertEqual(self.project_id, limits[0]['project_id'])

        r = self.get(
            '/limits', expected_status=http.client.OK,
            auth=self.build_authentication_request(
                user_id=self.user['id'], password=self.user['password'],
                project_id=self.project_2_id))
        limits = r.result['limits']
        self.assertEqual(1, len(limits))
        self.assertEqual(self.project_2_id, limits[0]['project_id'])

        # any project user can filter by their own project
        r = self.get(
            '/limits?project_id=%s' % self.project_id,
            expected_status=http.client.OK)
        limits = r.result['limits']
        self.assertEqual(1, len(limits))
        self.assertEqual(self.project_id, limits[0]['project_id'])

        # a system scoped request can specify the project_id filter
        r = self.get(
            '/limits?project_id=%s' % self.project_id,
            expected_status=http.client.OK,
            token=self.system_admin_token
        )
        limits = r.result['limits']
        self.assertEqual(1, len(limits))
        self.assertEqual(self.project_id, limits[0]['project_id'])

    def test_list_limit_with_domain_id_filter(self):
        # create two limit in different domains for test.
        ref1 = unit.new_limit_ref(domain_id=self.domain_id,
                                  service_id=self.service_id,
                                  region_id=self.region_id,
                                  resource_name='volume')
        ref2 = unit.new_limit_ref(domain_id=self.domain_2_id,
                                  service_id=self.service_id2,
                                  resource_name='snapshot')
        self.post(
            '/limits',
            body={'limits': [ref1, ref2]},
            token=self.system_admin_token,
            expected_status=http.client.CREATED)

        # non system scoped request will get the limits in its domain.
        r = self.get(
            '/limits', expected_status=http.client.OK,
            auth=self.build_authentication_request(
                user_id=self.user['id'], password=self.user['password'],
                domain_id=self.domain_id))
        limits = r.result['limits']
        self.assertEqual(1, len(limits))
        self.assertEqual(self.domain_id, limits[0]['domain_id'])

        r = self.get(
            '/limits', expected_status=http.client.OK,
            auth=self.build_authentication_request(
                user_id=self.user['id'], password=self.user['password'],
                domain_id=self.domain_2_id))
        limits = r.result['limits']
        self.assertEqual(1, len(limits))
        self.assertEqual(self.domain_2_id, limits[0]['domain_id'])

        # if non system scoped request contain domain_id filter, keystone
        # will return an empty list.
        r = self.get(
            '/limits?domain_id=%s' % self.domain_id,
            expected_status=http.client.OK)
        limits = r.result['limits']
        self.assertEqual(0, len(limits))

        # a system scoped request can specify the domain_id filter
        r = self.get(
            '/limits?domain_id=%s' % self.domain_id,
            expected_status=http.client.OK,
            auth=self.build_authentication_request(
                user_id=self.user['id'], password=self.user['password'],
                system=True))
        limits = r.result['limits']
        self.assertEqual(1, len(limits))
        self.assertEqual(self.domain_id, limits[0]['domain_id'])

    def test_show_project_limit(self):
        ref1 = unit.new_limit_ref(project_id=self.project_id,
                                  service_id=self.service_id,
                                  region_id=self.region_id,
                                  resource_name='volume')
        ref2 = unit.new_limit_ref(project_id=self.project_id,
                                  service_id=self.service_id2,
                                  resource_name='snapshot')
        r = self.post(
            '/limits',
            body={'limits': [ref1, ref2]},
            token=self.system_admin_token,
            expected_status=http.client.CREATED)
        if r.result['limits'][0]['resource_name'] == 'volume':
            id1 = r.result['limits'][0]['id']
        else:
            id1 = r.result['limits'][1]['id']
        self.get('/limits/fake_id',
                 token=self.system_admin_token,
                 expected_status=http.client.NOT_FOUND)
        r = self.get('/limits/%s' % id1,
                     expected_status=http.client.OK)
        limit = r.result['limit']
        self.assertIsNone(limit['domain_id'])
        for key in ['service_id', 'region_id', 'resource_name',
                    'resource_limit', 'description', 'project_id']:
            self.assertEqual(limit[key], ref1[key])

    def test_show_domain_limit(self):
        ref1 = unit.new_limit_ref(domain_id=self.domain_id,
                                  service_id=self.service_id2,
                                  resource_name='snapshot')
        r = self.post(
            '/limits',
            body={'limits': [ref1]},
            token=self.system_admin_token,
            expected_status=http.client.CREATED)
        id1 = r.result['limits'][0]['id']

        r = self.get('/limits/%s' % id1,
                     expected_status=http.client.OK,
                     auth=self.build_authentication_request(
                         user_id=self.user['id'],
                         password=self.user['password'],
                         domain_id=self.domain_id))
        limit = r.result['limit']
        self.assertIsNone(limit['project_id'])
        self.assertIsNone(limit['region_id'])
        for key in ['service_id', 'resource_name', 'resource_limit',
                    'description', 'domain_id']:
            self.assertEqual(limit[key], ref1[key])

    def test_delete_limit(self):
        ref1 = unit.new_limit_ref(project_id=self.project_id,
                                  service_id=self.service_id,
                                  region_id=self.region_id,
                                  resource_name='volume')
        ref2 = unit.new_limit_ref(project_id=self.project_id,
                                  service_id=self.service_id2,
                                  resource_name='snapshot')
        r = self.post(
            '/limits',
            body={'limits': [ref1, ref2]},
            token=self.system_admin_token,
            expected_status=http.client.CREATED)
        id1 = r.result['limits'][0]['id']
        self.delete('/limits/%s' % id1,
                    token=self.system_admin_token,
                    expected_status=http.client.NO_CONTENT)
        self.delete('/limits/fake_id',
                    token=self.system_admin_token,
                    expected_status=http.client.NOT_FOUND)
        r = self.get(
            '/limits',
            token=self.system_admin_token,
            expected_status=http.client.OK)
        limits = r.result['limits']
        self.assertEqual(len(limits), 1)


class StrictTwoLevelLimitsTestCase(LimitsTestCase):

    def setUp(self):
        super(StrictTwoLevelLimitsTestCase, self).setUp()
        # Most of these tests require system-scoped tokens. Let's have one on
        # hand so that we can use it in tests when we need it.
        PROVIDERS.assignment_api.create_system_grant_for_user(
            self.user_id, self.role_id
        )
        self.system_admin_token = self.get_system_scoped_token()

        # create two hierarchical projects trees for test. The first level is
        # domain.
        #   A        D
        #  / \      / \
        # B   C    E   F
        domain_ref = {'domain': {'name': 'A', 'enabled': True}}
        response = self.post('/domains', body=domain_ref)
        self.domain_A = response.json_body['domain']
        project_ref = {'project': {'name': 'B', 'enabled': True,
                                   'domain_id': self.domain_A['id']}}
        response = self.post('/projects', body=project_ref)
        self.project_B = response.json_body['project']
        project_ref = {'project': {'name': 'C', 'enabled': True,
                                   'domain_id': self.domain_A['id']}}
        response = self.post('/projects', body=project_ref)
        self.project_C = response.json_body['project']

        domain_ref = {'domain': {'name': 'D', 'enabled': True}}
        response = self.post('/domains', body=domain_ref)
        self.domain_D = response.json_body['domain']
        project_ref = {'project': {'name': 'E', 'enabled': True,
                                   'domain_id': self.domain_D['id']}}
        response = self.post('/projects', body=project_ref)
        self.project_E = response.json_body['project']
        project_ref = {'project': {'name': 'F', 'enabled': True,
                                   'domain_id': self.domain_D['id']}}
        response = self.post('/projects', body=project_ref)
        self.project_F = response.json_body['project']

    def config_overrides(self):
        super(StrictTwoLevelLimitsTestCase, self).config_overrides()
        self.config_fixture.config(group='unified_limit',
                                   enforcement_model='strict_two_level')

    def test_create_child_limit(self):
        # when A is 20, success to create B to 15, C to 18.
        #     A,20             A,20
        #    / \      -->     / \
        #   B   C           B,15 C,18
        ref = unit.new_limit_ref(domain_id=self.domain_A['id'],
                                 service_id=self.service_id,
                                 region_id=self.region_id,
                                 resource_name='volume',
                                 resource_limit=20)
        self.post(
            '/limits',
            body={'limits': [ref]},
            token=self.system_admin_token,
            expected_status=http.client.CREATED)

        ref = unit.new_limit_ref(project_id=self.project_B['id'],
                                 service_id=self.service_id,
                                 region_id=self.region_id,
                                 resource_name='volume',
                                 resource_limit=15)
        self.post(
            '/limits',
            body={'limits': [ref]},
            token=self.system_admin_token,
            expected_status=http.client.CREATED)

        ref = unit.new_limit_ref(project_id=self.project_C['id'],
                                 service_id=self.service_id,
                                 region_id=self.region_id,
                                 resource_name='volume',
                                 resource_limit=18)
        self.post(
            '/limits',
            body={'limits': [ref]},
            token=self.system_admin_token,
            expected_status=http.client.CREATED)

    def test_create_child_limit_break_hierarchical_tree(self):
        # when A is 20, success to create B to 15, but fail to create C to 21.
        #     A,20             A,20
        #    / \      -->     / \
        #   B   C           B,15 C
        #
        #     A,20              A,20
        #    / \      -/->      / \
        #  B,15 C            B,15 C,21
        ref = unit.new_limit_ref(domain_id=self.domain_A['id'],
                                 service_id=self.service_id,
                                 region_id=self.region_id,
                                 resource_name='volume',
                                 resource_limit=20)
        self.post(
            '/limits',
            body={'limits': [ref]},
            token=self.system_admin_token,
            expected_status=http.client.CREATED)

        ref = unit.new_limit_ref(project_id=self.project_B['id'],
                                 service_id=self.service_id,
                                 region_id=self.region_id,
                                 resource_name='volume',
                                 resource_limit=15)
        self.post(
            '/limits',
            body={'limits': [ref]},
            token=self.system_admin_token,
            expected_status=http.client.CREATED)

        ref = unit.new_limit_ref(project_id=self.project_C['id'],
                                 service_id=self.service_id,
                                 region_id=self.region_id,
                                 resource_name='volume',
                                 resource_limit=21)
        self.post(
            '/limits',
            body={'limits': [ref]},
            token=self.system_admin_token,
            expected_status=http.client.FORBIDDEN)

    def test_create_child_with_default_parent(self):
        # If A is not set, the default value is 10 (from registered limit).
        # success to create B to 5, but fail to create C to 11.
        #     A(10)             A(10)
        #    / \      -->      / \
        #   B   C            B,5  C
        #
        #     A(10)             A(10)
        #    / \     -/->      / \
        #   B,5   C          B,5  C,11
        ref = unit.new_limit_ref(project_id=self.project_B['id'],
                                 service_id=self.service_id,
                                 region_id=self.region_id,
                                 resource_name='volume',
                                 resource_limit=5)
        self.post(
            '/limits',
            body={'limits': [ref]},
            token=self.system_admin_token,
            expected_status=http.client.CREATED)

        ref = unit.new_limit_ref(project_id=self.project_C['id'],
                                 service_id=self.service_id,
                                 region_id=self.region_id,
                                 resource_name='volume',
                                 resource_limit=11)
        self.post(
            '/limits',
            body={'limits': [ref]},
            token=self.system_admin_token,
            expected_status=http.client.FORBIDDEN)

    def test_create_parent_limit(self):
        # When B is 9 , success to set A to 12
        #     A              A,12
        #    / \    -->      / \
        #  B,9  C          B,9  C
        ref = unit.new_limit_ref(project_id=self.project_B['id'],
                                 service_id=self.service_id,
                                 region_id=self.region_id,
                                 resource_name='volume',
                                 resource_limit=9)
        self.post(
            '/limits',
            body={'limits': [ref]},
            token=self.system_admin_token,
            expected_status=http.client.CREATED)

        ref = unit.new_limit_ref(domain_id=self.domain_A['id'],
                                 service_id=self.service_id,
                                 region_id=self.region_id,
                                 resource_name='volume',
                                 resource_limit=12)
        self.post(
            '/limits',
            body={'limits': [ref]},
            token=self.system_admin_token,
            expected_status=http.client.CREATED)

    def test_create_parent_limit_break_hierarchical_tree(self):
        # When B is 9 , fail to set A to 8
        #     A              A,8
        #    / \    -/->     / \
        #  B,9  C          B,9  C
        ref = unit.new_limit_ref(project_id=self.project_B['id'],
                                 service_id=self.service_id,
                                 region_id=self.region_id,
                                 resource_name='volume',
                                 resource_limit=9)
        self.post(
            '/limits',
            body={'limits': [ref]},
            token=self.system_admin_token,
            expected_status=http.client.CREATED)

        ref = unit.new_limit_ref(domain_id=self.domain_A['id'],
                                 service_id=self.service_id,
                                 region_id=self.region_id,
                                 resource_name='volume',
                                 resource_limit=8)
        self.post(
            '/limits',
            body={'limits': [ref]},
            token=self.system_admin_token,
            expected_status=http.client.FORBIDDEN)

    def test_create_multi_limits(self):
        # success to create a tree in one request like:
        #    A,12         D,9
        #    / \          / \
        #  B,9  C,5    E,5   F,4
        ref_A = unit.new_limit_ref(domain_id=self.domain_A['id'],
                                   service_id=self.service_id,
                                   region_id=self.region_id,
                                   resource_name='volume',
                                   resource_limit=12)
        ref_B = unit.new_limit_ref(project_id=self.project_B['id'],
                                   service_id=self.service_id,
                                   region_id=self.region_id,
                                   resource_name='volume',
                                   resource_limit=9)
        ref_C = unit.new_limit_ref(project_id=self.project_C['id'],
                                   service_id=self.service_id,
                                   region_id=self.region_id,
                                   resource_name='volume',
                                   resource_limit=5)
        ref_D = unit.new_limit_ref(domain_id=self.domain_D['id'],
                                   service_id=self.service_id,
                                   region_id=self.region_id,
                                   resource_name='volume',
                                   resource_limit=9)
        ref_E = unit.new_limit_ref(project_id=self.project_E['id'],
                                   service_id=self.service_id,
                                   region_id=self.region_id,
                                   resource_name='volume',
                                   resource_limit=5)
        ref_F = unit.new_limit_ref(project_id=self.project_F['id'],
                                   service_id=self.service_id,
                                   region_id=self.region_id,
                                   resource_name='volume',
                                   resource_limit=4)
        self.post(
            '/limits',
            body={'limits': [ref_A, ref_B, ref_C, ref_D, ref_E, ref_F]},
            token=self.system_admin_token,
            expected_status=http.client.CREATED)

    def test_create_multi_limits_invalid_input(self):
        # fail to create a tree in one request like:
        #    A,12         D,9
        #    / \          / \
        #  B,9  C,5    E,5   F,10
        # because F will break the second limit tree.
        ref_A = unit.new_limit_ref(domain_id=self.domain_A['id'],
                                   service_id=self.service_id,
                                   region_id=self.region_id,
                                   resource_name='volume',
                                   resource_limit=12)
        ref_B = unit.new_limit_ref(project_id=self.project_B['id'],
                                   service_id=self.service_id,
                                   region_id=self.region_id,
                                   resource_name='volume',
                                   resource_limit=9)
        ref_C = unit.new_limit_ref(project_id=self.project_C['id'],
                                   service_id=self.service_id,
                                   region_id=self.region_id,
                                   resource_name='volume',
                                   resource_limit=5)
        ref_D = unit.new_limit_ref(domain_id=self.domain_D['id'],
                                   service_id=self.service_id,
                                   region_id=self.region_id,
                                   resource_name='volume',
                                   resource_limit=9)
        ref_E = unit.new_limit_ref(project_id=self.project_E['id'],
                                   service_id=self.service_id,
                                   region_id=self.region_id,
                                   resource_name='volume',
                                   resource_limit=5)
        ref_F = unit.new_limit_ref(project_id=self.project_F['id'],
                                   service_id=self.service_id,
                                   region_id=self.region_id,
                                   resource_name='volume',
                                   resource_limit=10)
        self.post(
            '/limits',
            body={'limits': [ref_A, ref_B, ref_C, ref_D, ref_E, ref_F]},
            token=self.system_admin_token,
            expected_status=http.client.FORBIDDEN)

    def test_create_multi_limits_break_hierarchical_tree(self):
        # when there is some hierarchical_trees already like:
        #    A,12          D
        #    / \          / \
        #  B,9  C       E,5   F
        # fail to set C to 5 and D to 4 in one request like:
        #    A,12         D,4
        #    / \          / \
        #  B,9  C,5    E,5   F
        # because D will break the second limit tree.
        ref_A = unit.new_limit_ref(domain_id=self.domain_A['id'],
                                   service_id=self.service_id,
                                   region_id=self.region_id,
                                   resource_name='volume',
                                   resource_limit=12)
        ref_B = unit.new_limit_ref(project_id=self.project_B['id'],
                                   service_id=self.service_id,
                                   region_id=self.region_id,
                                   resource_name='volume',
                                   resource_limit=9)
        ref_E = unit.new_limit_ref(project_id=self.project_E['id'],
                                   service_id=self.service_id,
                                   region_id=self.region_id,
                                   resource_name='volume',
                                   resource_limit=5)
        self.post(
            '/limits',
            body={'limits': [ref_A, ref_B, ref_E]},
            token=self.system_admin_token,
            expected_status=http.client.CREATED)

        ref_C = unit.new_limit_ref(project_id=self.project_C['id'],
                                   service_id=self.service_id,
                                   region_id=self.region_id,
                                   resource_name='volume',
                                   resource_limit=5)
        ref_D = unit.new_limit_ref(domain_id=self.domain_D['id'],
                                   service_id=self.service_id,
                                   region_id=self.region_id,
                                   resource_name='volume',
                                   resource_limit=4)
        self.post(
            '/limits',
            body={'limits': [ref_C, ref_D]},
            token=self.system_admin_token,
            expected_status=http.client.FORBIDDEN)

    def test_update_child_limit(self):
        # Success to update C to 9
        #     A,10             A,10
        #    / \      -->     / \
        #  B,6  C,7         B,6  C,9
        ref_A = unit.new_limit_ref(domain_id=self.domain_A['id'],
                                   service_id=self.service_id,
                                   region_id=self.region_id,
                                   resource_name='volume',
                                   resource_limit=10)
        ref_B = unit.new_limit_ref(project_id=self.project_B['id'],
                                   service_id=self.service_id,
                                   region_id=self.region_id,
                                   resource_name='volume',
                                   resource_limit=6)
        ref_C = unit.new_limit_ref(project_id=self.project_C['id'],
                                   service_id=self.service_id,
                                   region_id=self.region_id,
                                   resource_name='volume',
                                   resource_limit=7)
        self.post(
            '/limits',
            body={'limits': [ref_A, ref_B]},
            token=self.system_admin_token,
            expected_status=http.client.CREATED)
        r = self.post(
            '/limits',
            body={'limits': [ref_C]},
            token=self.system_admin_token,
            expected_status=http.client.CREATED)

        update_dict = {'resource_limit': 9}
        self.patch(
            '/limits/%s' % r.result['limits'][0]['id'],
            body={'limit': update_dict},
            token=self.system_admin_token,
            expected_status=http.client.OK)

    def test_update_child_limit_break_hierarchical_tree(self):
        # Fail to update C to 11
        #     A,10             A,10
        #    / \      -/->     / \
        #  B,6  C,7         B,6  C,11
        ref_A = unit.new_limit_ref(domain_id=self.domain_A['id'],
                                   service_id=self.service_id,
                                   region_id=self.region_id,
                                   resource_name='volume',
                                   resource_limit=10)
        ref_B = unit.new_limit_ref(project_id=self.project_B['id'],
                                   service_id=self.service_id,
                                   region_id=self.region_id,
                                   resource_name='volume',
                                   resource_limit=6)
        ref_C = unit.new_limit_ref(project_id=self.project_C['id'],
                                   service_id=self.service_id,
                                   region_id=self.region_id,
                                   resource_name='volume',
                                   resource_limit=7)
        self.post(
            '/limits',
            body={'limits': [ref_A, ref_B]},
            token=self.system_admin_token,
            expected_status=http.client.CREATED)
        r = self.post(
            '/limits',
            body={'limits': [ref_C]},
            token=self.system_admin_token,
            expected_status=http.client.CREATED)

        update_dict = {'resource_limit': 11}
        self.patch(
            '/limits/%s' % r.result['limits'][0]['id'],
            body={'limit': update_dict},
            token=self.system_admin_token,
            expected_status=http.client.FORBIDDEN)

    def test_update_child_limit_with_default_parent(self):
        # If A is not set, the default value is 10 (from registered limit).
        # Success to update C to 9 but fail to update C to 11
        #     A,(10)           A,(10)
        #    / \      -->     / \
        #   B,  C,7          B  C,9
        #
        #     A,(10)            A,(10)
        #    / \      -/->     / \
        #   B,  C,7           B  C,11
        ref_C = unit.new_limit_ref(project_id=self.project_C['id'],
                                   service_id=self.service_id,
                                   region_id=self.region_id,
                                   resource_name='volume',
                                   resource_limit=7)
        r = self.post(
            '/limits',
            body={'limits': [ref_C]},
            token=self.system_admin_token,
            expected_status=http.client.CREATED)

        update_dict = {'resource_limit': 9}
        self.patch(
            '/limits/%s' % r.result['limits'][0]['id'],
            body={'limit': update_dict},
            token=self.system_admin_token,
            expected_status=http.client.OK)

        update_dict = {'resource_limit': 11}
        self.patch(
            '/limits/%s' % r.result['limits'][0]['id'],
            body={'limit': update_dict},
            token=self.system_admin_token,
            expected_status=http.client.FORBIDDEN)

    def test_update_parent_limit(self):
        # Success to update A to 8
        #     A,10             A,8
        #    / \      -->     / \
        #  B,6  C,7         B,6  C,7
        ref_A = unit.new_limit_ref(domain_id=self.domain_A['id'],
                                   service_id=self.service_id,
                                   region_id=self.region_id,
                                   resource_name='volume',
                                   resource_limit=10)
        ref_B = unit.new_limit_ref(project_id=self.project_B['id'],
                                   service_id=self.service_id,
                                   region_id=self.region_id,
                                   resource_name='volume',
                                   resource_limit=6)
        ref_C = unit.new_limit_ref(project_id=self.project_C['id'],
                                   service_id=self.service_id,
                                   region_id=self.region_id,
                                   resource_name='volume',
                                   resource_limit=7)
        r = self.post(
            '/limits',
            body={'limits': [ref_A]},
            token=self.system_admin_token,
            expected_status=http.client.CREATED)
        self.post(
            '/limits',
            body={'limits': [ref_B, ref_C]},
            token=self.system_admin_token,
            expected_status=http.client.CREATED)

        update_dict = {'resource_limit': 8}
        self.patch(
            '/limits/%s' % r.result['limits'][0]['id'],
            body={'limit': update_dict},
            token=self.system_admin_token,
            expected_status=http.client.OK)

    def test_update_parent_limit_break_hierarchical_tree(self):
        # Fail to update A to 6
        #     A,10             A,6
        #    / \      -/->     / \
        #  B,6  C,7         B,6  C,7
        ref_A = unit.new_limit_ref(domain_id=self.domain_A['id'],
                                   service_id=self.service_id,
                                   region_id=self.region_id,
                                   resource_name='volume',
                                   resource_limit=10)
        ref_B = unit.new_limit_ref(project_id=self.project_B['id'],
                                   service_id=self.service_id,
                                   region_id=self.region_id,
                                   resource_name='volume',
                                   resource_limit=6)
        ref_C = unit.new_limit_ref(project_id=self.project_C['id'],
                                   service_id=self.service_id,
                                   region_id=self.region_id,
                                   resource_name='volume',
                                   resource_limit=7)
        r = self.post(
            '/limits',
            body={'limits': [ref_A]},
            token=self.system_admin_token,
            expected_status=http.client.CREATED)
        self.post(
            '/limits',
            body={'limits': [ref_B, ref_C]},
            token=self.system_admin_token,
            expected_status=http.client.CREATED)

        update_dict = {'resource_limit': 6}
        self.patch(
            '/limits/%s' % r.result['limits'][0]['id'],
            body={'limit': update_dict},
            token=self.system_admin_token,
            expected_status=http.client.FORBIDDEN)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import datetime
import uuid

from oslo_utils import timeutils

from keystone.common import provider_api
from keystone import exception
from keystone.tests.unit import core

PROVIDERS = provider_api.ProviderAPIs


class TrustTests(object):
    def create_sample_trust(self, new_id, remaining_uses=None):
        self.trustor = self.user_foo
        self.trustee = self.user_two
        expires_at = datetime.datetime.utcnow().replace(year=2032)
        trust_data = (PROVIDERS.trust_api.create_trust
                      (new_id,
                       {'trustor_user_id': self.trustor['id'],
                        'trustee_user_id': self.user_two['id'],
                        'project_id': self.project_bar['id'],
                        'expires_at': expires_at,
                        'impersonation': True,
                        'remaining_uses': remaining_uses},
                       roles=[{"id": "member"},
                              {"id": "other"},
                              {"id": "browser"}]))
        return trust_data

    def test_delete_trust(self):
        new_id = uuid.uuid4().hex
        trust_data = self.create_sample_trust(new_id)
        trust_id = trust_data['id']
        self.assertIsNotNone(trust_data)
        trust_data = PROVIDERS.trust_api.get_trust(trust_id)
        self.assertEqual(new_id, trust_data['id'])
        PROVIDERS.trust_api.delete_trust(trust_id)
        self.assertRaises(exception.TrustNotFound,
                          PROVIDERS.trust_api.get_trust,
                          trust_id)

    def test_delete_trust_not_found(self):
        trust_id = uuid.uuid4().hex
        self.assertRaises(exception.TrustNotFound,
                          PROVIDERS.trust_api.delete_trust,
                          trust_id)

    def test_get_trust(self):
        new_id = uuid.uuid4().hex
        trust_data = self.create_sample_trust(new_id)
        trust_id = trust_data['id']
        self.assertIsNotNone(trust_data)
        trust_data = PROVIDERS.trust_api.get_trust(trust_id)
        self.assertEqual(new_id, trust_data['id'])
        PROVIDERS.trust_api.delete_trust(trust_data['id'])

    def test_get_deleted_trust(self):
        new_id = uuid.uuid4().hex
        trust_data = self.create_sample_trust(new_id)
        self.assertIsNotNone(trust_data)
        self.assertIsNone(trust_data['deleted_at'])
        PROVIDERS.trust_api.delete_trust(new_id)
        self.assertRaises(exception.TrustNotFound,
                          PROVIDERS.trust_api.get_trust,
                          new_id)
        deleted_trust = PROVIDERS.trust_api.get_trust(
            trust_data['id'], deleted=True
        )
        self.assertEqual(trust_data['id'], deleted_trust['id'])
        self.assertIsNotNone(deleted_trust.get('deleted_at'))

    def test_create_trust(self):
        new_id = uuid.uuid4().hex
        trust_data = self.create_sample_trust(new_id)

        self.assertEqual(new_id, trust_data['id'])
        self.assertEqual(self.trustee['id'], trust_data['trustee_user_id'])
        self.assertEqual(self.trustor['id'], trust_data['trustor_user_id'])
        self.assertGreater(timeutils.normalize_time(trust_data['expires_at']),
                           timeutils.utcnow())

        self.assertEqual([{'id': 'member'},
                          {'id': 'other'},
                          {'id': 'browser'}], trust_data['roles'])

    def test_list_trust_by_trustee(self):
        for i in range(3):
            self.create_sample_trust(uuid.uuid4().hex)
        trusts = PROVIDERS.trust_api.list_trusts_for_trustee(
            self.trustee['id']
        )
        self.assertEqual(3, len(trusts))
        self.assertEqual(trusts[0]["trustee_user_id"], self.trustee['id'])
        trusts = PROVIDERS.trust_api.list_trusts_for_trustee(
            self.trustor['id']
        )
        self.assertEqual(0, len(trusts))

    def test_list_trust_by_trustor(self):
        for i in range(3):
            self.create_sample_trust(uuid.uuid4().hex)
        trusts = PROVIDERS.trust_api.list_trusts_for_trustor(
            self.trustor['id']
        )
        self.assertEqual(3, len(trusts))
        self.assertEqual(trusts[0]["trustor_user_id"], self.trustor['id'])
        trusts = PROVIDERS.trust_api.list_trusts_for_trustor(
            self.trustee['id']
        )
        self.assertEqual(0, len(trusts))

    def test_list_trusts(self):
        for i in range(3):
            self.create_sample_trust(uuid.uuid4().hex)
        trusts = PROVIDERS.trust_api.list_trusts()
        self.assertEqual(3, len(trusts))

    def test_trust_has_remaining_uses_positive(self):
        # create a trust with limited uses, check that we have uses left
        trust_data = self.create_sample_trust(uuid.uuid4().hex,
                                              remaining_uses=5)
        self.assertEqual(5, trust_data['remaining_uses'])
        # create a trust with unlimited uses, check that we have uses left
        trust_data = self.create_sample_trust(uuid.uuid4().hex)
        self.assertIsNone(trust_data['remaining_uses'])

    def test_trust_has_remaining_uses_negative(self):
        # try to create a trust with no remaining uses, check that it fails
        self.assertRaises(exception.ValidationError,
                          self.create_sample_trust,
                          uuid.uuid4().hex,
                          remaining_uses=0)
        # try to create a trust with negative remaining uses,
        # check that it fails
        self.assertRaises(exception.ValidationError,
                          self.create_sample_trust,
                          uuid.uuid4().hex,
                          remaining_uses=-12)

    def test_consume_use(self):
        # consume a trust repeatedly until it has no uses anymore
        trust_data = self.create_sample_trust(uuid.uuid4().hex,
                                              remaining_uses=2)
        PROVIDERS.trust_api.consume_use(trust_data['id'])
        t = PROVIDERS.trust_api.get_trust(trust_data['id'])
        self.assertEqual(1, t['remaining_uses'])
        PROVIDERS.trust_api.consume_use(trust_data['id'])
        # This was the last use, the trust isn't available anymore
        self.assertRaises(exception.TrustNotFound,
                          PROVIDERS.trust_api.get_trust,
                          trust_data['id'])

    def test_duplicate_trusts_not_allowed(self):
        self.trustor = self.user_foo
        self.trustee = self.user_two
        trust_data = {'trustor_user_id': self.trustor['id'],
                      'trustee_user_id': self.user_two['id'],
                      'project_id': self.project_bar['id'],
                      'expires_at': timeutils.parse_isotime(
                          '2032-02-18T18:10:00Z'),
                      'impersonation': True,
                      'remaining_uses': None}
        roles = [{"id": "member"},
                 {"id": "other"},
                 {"id": "browser"}]
        PROVIDERS.trust_api.create_trust(uuid.uuid4().hex, trust_data, roles)
        self.assertRaises(exception.Conflict,
                          PROVIDERS.trust_api.create_trust,
                          uuid.uuid4().hex,
                          trust_data,
                          roles)

    def test_flush_expired_trusts(self):
        roles = [{"id": "member"},
                 {"id": "other"},
                 {"id": "browser"}]
        trust_ref1 = core.new_trust_ref(
            self.user_foo['id'], self.user_two['id'],
            project_id=self.project_bar['id'])
        trust_ref1['expires_at'] = \
            timeutils.utcnow() - datetime.timedelta(minutes=1)
        trust_ref2 = core.new_trust_ref(
            self.user_foo['id'], self.user_two['id'],
            project_id=self.project_bar['id'])
        trust_ref2['expires_at'] = \
            timeutils.utcnow() + datetime.timedelta(minutes=1)

        trust_data = PROVIDERS.trust_api.create_trust(trust_ref1['id'],
                                                      trust_ref1, roles)
        self.assertIsNotNone(trust_data)
        trust_data = PROVIDERS.trust_api.create_trust(trust_ref2['id'],
                                                      trust_ref2, roles)
        self.assertIsNotNone(trust_data)

        PROVIDERS.trust_api.flush_expired_and_soft_deleted_trusts(
            date=datetime.datetime.utcnow())
        trusts = self.trust_api.list_trusts()
        self.assertEqual(len(trusts), 1)
        self.assertEqual(trust_ref2['id'], trusts[0]['id'])

    def test_flush_expired_trusts_with_all_id(self):
        roles = [{"id": "member"},
                 {"id": "other"},
                 {"id": "browser"}]
        trust_ref1 = core.new_trust_ref(
            self.user_foo['id'], self.user_foo['id'],
            project_id=self.project_bar['id'])
        trust_ref1['expires_at'] = \
            timeutils.utcnow() - datetime.timedelta(minutes=1)
        trust_ref2 = core.new_trust_ref(
            self.user_foo['id'], self.user_two['id'],
            project_id=self.project_bar['id'])
        trust_ref2['expires_at'] = \
            timeutils.utcnow() - datetime.timedelta(minutes=5)

        trust_data = PROVIDERS.trust_api.create_trust(trust_ref1['id'],
                                                      trust_ref1, roles)
        self.assertIsNotNone(trust_data)
        trust_data = PROVIDERS.trust_api.create_trust(trust_ref2['id'],
                                                      trust_ref2, roles)
        self.assertIsNotNone(trust_data)

        PROVIDERS.trust_api.flush_expired_and_soft_deleted_trusts(
            project_id=self.project_bar['id'],
            trustor_user_id=self.user_foo['id'],
            trustee_user_id=self.user_two['id'],
            date=datetime.datetime.utcnow())
        trusts = self.trust_api.list_trusts()
        self.assertEqual(len(trusts), 1)
        self.assertEqual(trust_ref1['id'], trusts[0]['id'])

    def test_flush_expired_trusts_with_no_project_id(self):
        roles = [{"id": "member"},
                 {"id": "other"},
                 {"id": "browser"}]
        trust_ref1 = core.new_trust_ref(
            self.user_foo['id'], self.user_two['id'],
            project_id=self.project_bar['id'])
        trust_ref1['expires_at'] = \
            timeutils.utcnow() - datetime.timedelta(minutes=1)
        trust_ref2 = core.new_trust_ref(
            self.user_foo['id'], self.user_two['id'],
            project_id=self.project_bar['id'])
        trust_ref2['expires_at'] = \
            timeutils.utcnow() + datetime.timedelta(minutes=1)

        trust_data = PROVIDERS.trust_api.create_trust(trust_ref1['id'],
                                                      trust_ref1, roles)
        self.assertIsNotNone(trust_data)
        trust_data = PROVIDERS.trust_api.create_trust(trust_ref2['id'],
                                                      trust_ref2, roles)
        self.assertIsNotNone(trust_data)

        PROVIDERS.trust_api.flush_expired_and_soft_deleted_trusts(
            trustor_user_id=self.user_foo['id'],
            trustee_user_id=self.user_two['id'],
            date=datetime.datetime.utcnow())
        trusts = self.trust_api.list_trusts()
        self.assertEqual(len(trusts), 1)
        self.assertEqual(trust_ref2['id'], trusts[0]['id'])

    def test_flush_expired_trusts_with_no_trustor_id(self):
        roles = [{"id": "member"},
                 {"id": "other"},
                 {"id": "browser"}]
        trust_ref1 = core.new_trust_ref(
            self.user_foo['id'], self.user_two['id'],
            project_id=self.project_bar['id'])
        trust_ref1['expires_at'] = \
            timeutils.utcnow() - datetime.timedelta(minutes=1)
        trust_ref2 = core.new_trust_ref(
            self.user_foo['id'], self.user_two['id'],
            project_id=self.project_bar['id'])
        trust_ref2['expires_at'] = \
            timeutils.utcnow() + datetime.timedelta(minutes=1)

        trust_data = PROVIDERS.trust_api.create_trust(trust_ref1['id'],
                                                      trust_ref1, roles)
        self.assertIsNotNone(trust_data)
        trust_data = PROVIDERS.trust_api.create_trust(trust_ref2['id'],
                                                      trust_ref2, roles)
        self.assertIsNotNone(trust_data)

        PROVIDERS.trust_api.flush_expired_and_soft_deleted_trusts(
            project_id=self.project_bar['id'],
            trustee_user_id=self.user_two['id'],
            date=datetime.datetime.utcnow())
        trusts = self.trust_api.list_trusts()
        self.assertEqual(len(trusts), 1)
        self.assertEqual(trust_ref2['id'], trusts[0]['id'])

    def test_flush_expired_trusts_with_no_trustee_id(self):
        roles = [{"id": "member"},
                 {"id": "other"},
                 {"id": "browser"}]
        trust_ref1 = core.new_trust_ref(
            self.user_foo['id'], self.user_two['id'],
            project_id=self.project_bar['id'])
        trust_ref1['expires_at'] = \
            timeutils.utcnow() - datetime.timedelta(minutes=1)
        trust_ref2 = core.new_trust_ref(
            self.user_foo['id'], self.user_two['id'],
            project_id=self.project_bar['id'])
        trust_ref2['expires_at'] = \
            timeutils.utcnow() + datetime.timedelta(minutes=1)

        trust_data = PROVIDERS.trust_api.create_trust(trust_ref1['id'],
                                                      trust_ref1, roles)
        self.assertIsNotNone(trust_data)
        trust_data = PROVIDERS.trust_api.create_trust(trust_ref2['id'],
                                                      trust_ref2, roles)
        self.assertIsNotNone(trust_data)

        PROVIDERS.trust_api.flush_expired_and_soft_deleted_trusts(
            project_id=self.project_bar['id'],
            trustor_user_id=self.user_foo['id'],
            date=datetime.datetime.utcnow())
        trusts = self.trust_api.list_trusts()
        self.assertEqual(len(trusts), 1)
        self.assertEqual(trust_ref2['id'], trusts[0]['id'])

    def test_flush_expired_trusts_with_project_id(self):
        roles = [{"id": "member"},
                 {"id": "other"},
                 {"id": "browser"}]
        trust_ref1 = core.new_trust_ref(
            self.user_foo['id'], self.user_two['id'],
            project_id=self.project_bar['id'])
        trust_ref1['expires_at'] = \
            timeutils.utcnow() - datetime.timedelta(minutes=1)
        trust_ref2 = core.new_trust_ref(
            self.user_foo['id'], self.user_two['id'],
            project_id=self.user_foo['id'])
        trust_ref2['expires_at'] = \
            timeutils.utcnow() - datetime.timedelta(minutes=5)

        trust_data = PROVIDERS.trust_api.create_trust(trust_ref1['id'],
                                                      trust_ref1, roles)
        self.assertIsNotNone(trust_data)
        trust_data = PROVIDERS.trust_api.create_trust(trust_ref2['id'],
                                                      trust_ref2, roles)
        self.assertIsNotNone(trust_data)

        PROVIDERS.trust_api.flush_expired_and_soft_deleted_trusts(
            project_id=self.project_bar['id'],
            date=datetime.datetime.utcnow())
        trusts = self.trust_api.list_trusts()
        self.assertEqual(len(trusts), 1)
        self.assertEqual(trust_ref2['id'], trusts[0]['id'])

    def test_flush_expired_trusts_with_trustee_id(self):
        roles = [{"id": "member"},
                 {"id": "other"},
                 {"id": "browser"}]
        trust_ref1 = core.new_trust_ref(
            self.user_foo['id'], self.user_two['id'],
            project_id=self.project_bar['id'])
        trust_ref1['expires_at'] = \
            timeutils.utcnow() - datetime.timedelta(minutes=1)
        trust_ref2 = core.new_trust_ref(
            self.user_foo['id'], self.user_foo['id'],
            project_id=self.project_bar['id'])
        trust_ref2['expires_at'] = \
            timeutils.utcnow() - datetime.timedelta(minutes=5)

        trust_data = PROVIDERS.trust_api.create_trust(trust_ref1['id'],
                                                      trust_ref1, roles)
        self.assertIsNotNone(trust_data)
        trust_data = PROVIDERS.trust_api.create_trust(trust_ref2['id'],
                                                      trust_ref2, roles)
        self.assertIsNotNone(trust_data)
        PROVIDERS.trust_api.flush_expired_and_soft_deleted_trusts(
            trustee_user_id=self.user_two['id'],
            date=datetime.datetime.utcnow())
        trusts = self.trust_api.list_trusts()
        self.assertEqual(len(trusts), 1)
        self.assertEqual(trust_ref2['id'], trusts[0]['id'])

    def test_flush_expired_trusts_with_trustor_id(self):
        roles = [{"id": "member"},
                 {"id": "other"},
                 {"id": "browser"}]
        trust_ref1 = core.new_trust_ref(
            self.user_foo['id'], self.user_two['id'],
            project_id=self.project_bar['id'])
        trust_ref1['expires_at'] = \
            timeutils.utcnow() - datetime.timedelta(minutes=1)
        trust_ref2 = core.new_trust_ref(
            self.user_two['id'], self.user_two['id'],
            project_id=self.project_bar['id'])
        trust_ref2['expires_at'] = \
            timeutils.utcnow() - datetime.timedelta(minutes=5)

        trust_data = PROVIDERS.trust_api.create_trust(trust_ref1['id'],
                                                      trust_ref1, roles)
        self.assertIsNotNone(trust_data)
        trust_data = PROVIDERS.trust_api.create_trust(trust_ref2['id'],
                                                      trust_ref2, roles)
        self.assertIsNotNone(trust_data)

        PROVIDERS.trust_api.flush_expired_and_soft_deleted_trusts(
            trustor_user_id=self.user_foo['id'],
            date=datetime.datetime.utcnow())
        trusts = self.trust_api.list_trusts()
        self.assertEqual(len(trusts), 1)
        self.assertEqual(trust_ref2['id'], trusts[0]['id'])

    def test_non_expired_soft_deleted_trusts(self):
        roles = [{"id": "member"},
                 {"id": "other"},
                 {"id": "browser"}]
        trust_ref1 = core.new_trust_ref(
            self.user_foo['id'], self.user_two['id'],
            project_id=self.project_bar['id'])
        trust_ref1['expires_at'] = \
            timeutils.utcnow() + datetime.timedelta(minutes=10)
        trust_ref2 = core.new_trust_ref(
            self.user_two['id'], self.user_two['id'],
            project_id=self.project_bar['id'])
        trust_ref2['expires_at'] = \
            timeutils.utcnow() + datetime.timedelta(minutes=5)

        trust_data = PROVIDERS.trust_api.create_trust(trust_ref1['id'],
                                                      trust_ref1, roles)
        self.assertIsNotNone(trust_data)
        trust_data = PROVIDERS.trust_api.create_trust(trust_ref2['id'],
                                                      trust_ref2, roles)
        self.assertIsNotNone(trust_data)
        PROVIDERS.trust_api.delete_trust(trust_ref2['id'])

        PROVIDERS.trust_api.flush_expired_and_soft_deleted_trusts(
            date=datetime.datetime.utcnow())
        trusts = self.trust_api.list_trusts()
        self.assertEqual(len(trusts), 1)
        self.assertEqual(trust_ref1['id'], trusts[0]['id'])

    def test_non_expired_non_deleted_trusts(self):
        roles = [{"id": "member"},
                 {"id": "other"},
                 {"id": "browser"}]
        trust_ref1 = core.new_trust_ref(
            self.user_foo['id'], self.user_two['id'],
            project_id=self.project_bar['id'])
        trust_ref1['expires_at'] = \
            timeutils.utcnow() + datetime.timedelta(minutes=10)
        trust_ref2 = core.new_trust_ref(
            self.user_two['id'], self.user_two['id'],
            project_id=self.project_bar['id'])
        trust_ref2['expires_at'] = \
            timeutils.utcnow() + datetime.timedelta(minutes=5)
        trust_ref3 = core.new_trust_ref(
            self.user_two['id'], self.user_foo['id'],
            project_id=self.project_bar['id'])
        trust_ref3['expires_at'] = None

        trust_data = PROVIDERS.trust_api.create_trust(trust_ref1['id'],
                                                      trust_ref1, roles)
        self.assertIsNotNone(trust_data)
        trust_data = PROVIDERS.trust_api.create_trust(trust_ref2['id'],
                                                      trust_ref2, roles)
        self.assertIsNotNone(trust_data)
        PROVIDERS.trust_api.delete_trust(trust_ref2['id'])
        trust_data = PROVIDERS.trust_api.create_trust(trust_ref3['id'],
                                                      trust_ref3, roles)
        self.assertIsNotNone(trust_data)

        PROVIDERS.trust_api.flush_expired_and_soft_deleted_trusts(
            date=datetime.datetime.utcnow())
        trusts = self.trust_api.list_trusts()
        self.assertEqual(len(trusts), 2)

    def test_flush_expired_trusts_with_date(self):
        roles = [{"id": "member"},
                 {"id": "other"},
                 {"id": "browser"}]
        trust_ref1 = core.new_trust_ref(
            self.user_foo['id'], self.user_two['id'],
            project_id=self.project_bar['id'])
        trust_ref1['expires_at'] = \
            timeutils.utcnow() + datetime.timedelta(minutes=10)
        trust_ref2 = core.new_trust_ref(
            self.user_two['id'], self.user_two['id'],
            project_id=self.project_bar['id'])
        trust_ref2['expires_at'] = \
            timeutils.utcnow() + datetime.timedelta(minutes=30)
        trust_ref3 = core.new_trust_ref(
            self.user_two['id'], self.user_foo['id'],
            project_id=self.project_bar['id'])
        trust_ref3['expires_at'] = \
            timeutils.utcnow() - datetime.timedelta(minutes=30)

        trust_data = PROVIDERS.trust_api.create_trust(trust_ref1['id'],
                                                      trust_ref1, roles)
        self.assertIsNotNone(trust_data)
        trust_data = PROVIDERS.trust_api.create_trust(trust_ref2['id'],
                                                      trust_ref2, roles)
        self.assertIsNotNone(trust_data)
        trust_data = PROVIDERS.trust_api.create_trust(trust_ref3['id'],
                                                      trust_ref3, roles)
        self.assertIsNotNone(trust_data)
        fake_date = timeutils.utcnow() + datetime.timedelta(minutes=15)
        PROVIDERS.trust_api.flush_expired_and_soft_deleted_trusts(
            date=fake_date
        )
        trusts = self.trust_api.list_trusts()
        self.assertEqual(len(trusts), 1)
        self.assertEqual(trust_ref2['id'], trusts[0]['id'])



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# This file handles all flask-restful resources for /v3/OS-OAUTH1/

import flask
import flask_restful
import http.client
from oslo_log import log
from oslo_utils import timeutils
from urllib import parse as urlparse
from werkzeug import exceptions

from keystone.api._shared import json_home_relations
from keystone.common import authorization
from keystone.common import context
from keystone.common import provider_api
from keystone.common import rbac_enforcer
from keystone.common import validation
import keystone.conf
from keystone import exception
from keystone.i18n import _
from keystone import notifications
from keystone.oauth1 import core as oauth1
from keystone.oauth1 import schema
from keystone.oauth1 import validator
from keystone.server import flask as ks_flask


LOG = log.getLogger(__name__)
PROVIDERS = provider_api.ProviderAPIs
ENFORCER = rbac_enforcer.RBACEnforcer
CONF = keystone.conf.CONF


_build_resource_relation = json_home_relations.os_oauth1_resource_rel_func
_build_parameter_relation = json_home_relations.os_oauth1_parameter_rel_func

_ACCESS_TOKEN_ID_PARAMETER_RELATION = _build_parameter_relation(
    parameter_name='access_token_id')


def _normalize_role_list(authorize_roles):
    roles = set()
    for role in authorize_roles:
        if role.get('id'):
            roles.add(role['id'])
        else:
            roles.add(PROVIDERS.role_api.get_unique_role_by_name(
                role['name'])['id'])
    return roles


def _update_url_scheme():
    """Update request url scheme with base url scheme."""
    url = ks_flask.base_url()
    url_scheme = list(urlparse.urlparse(url))[0]
    req_url_list = list(urlparse.urlparse(flask.request.url))
    req_url_list[0] = url_scheme
    req_url = urlparse.urlunparse(req_url_list)
    return req_url


class _OAuth1ResourceBase(flask_restful.Resource):
    def get(self):
        # GET is not allowed, however flask restful doesn't handle "GET" not
        # being allowed cleanly. Here we explicitly mark is as not allowed. All
        # other methods not defined would raise a method NotAllowed error and
        # this would not be needed.
        raise exceptions.MethodNotAllowed(valid_methods=['POST'])


class ConsumerResource(ks_flask.ResourceBase):
    collection_key = 'consumers'
    member_key = 'consumer'
    api_prefix = '/OS-OAUTH1'
    json_home_resource_rel_func = _build_resource_relation
    json_home_parameter_rel_func = _build_parameter_relation

    def _list_consumers(self):
        ENFORCER.enforce_call(action='identity:list_consumers')
        return self.wrap_collection(PROVIDERS.oauth_api.list_consumers())

    def _get_consumer(self, consumer_id):
        ENFORCER.enforce_call(action='identity:get_consumer')
        return self.wrap_member(PROVIDERS.oauth_api.get_consumer(consumer_id))

    def get(self, consumer_id=None):
        if consumer_id is None:
            return self._list_consumers()
        return self._get_consumer(consumer_id)

    def post(self):
        ENFORCER.enforce_call(action='identity:create_consumer')
        consumer = (flask.request.get_json(force=True, silent=True) or {}).get(
            'consumer', {})
        consumer = self._normalize_dict(consumer)
        validation.lazy_validate(schema.consumer_create, consumer)
        consumer = self._assign_unique_id(consumer)
        ref = PROVIDERS.oauth_api.create_consumer(
            consumer, initiator=self.audit_initiator)
        return self.wrap_member(ref), http.client.CREATED

    def delete(self, consumer_id):
        ENFORCER.enforce_call(action='identity:delete_consumer')
        reason = (
            'Invalidating token cache because consumer %(consumer_id)s has '
            'been deleted. Authorization for users with OAuth tokens will be '
            'recalculated and enforced accordingly the next time they '
            'authenticate or validate a token.' %
            {'consumer_id': consumer_id}
        )
        notifications.invalidate_token_cache_notification(reason)
        PROVIDERS.oauth_api.delete_consumer(
            consumer_id, initiator=self.audit_initiator)
        return None, http.client.NO_CONTENT

    def patch(self, consumer_id):
        ENFORCER.enforce_call(action='identity:update_consumer')
        consumer = (flask.request.get_json(force=True, silent=True) or {}).get(
            'consumer', {})
        validation.lazy_validate(schema.consumer_update, consumer)
        consumer = self._normalize_dict(consumer)
        self._require_matching_id(consumer)
        ref = PROVIDERS.oauth_api.update_consumer(
            consumer_id, consumer, initiator=self.audit_initiator)
        return self.wrap_member(ref)


class RequestTokenResource(_OAuth1ResourceBase):
    @ks_flask.unenforced_api
    def post(self):
        oauth_headers = oauth1.get_oauth_headers(flask.request.headers)
        consumer_id = oauth_headers.get('oauth_consumer_key')
        requested_project_id = flask.request.headers.get(
            'Requested-Project-Id')

        if not consumer_id:
            raise exception.ValidationError(
                attribute='oauth_consumer_key', target='request')
        if not requested_project_id:
            raise exception.ValidationError(
                attribute='Requested-Project-Id', target='request')

        # NOTE(stevemar): Ensure consumer and requested project exist
        PROVIDERS.resource_api.get_project(requested_project_id)
        PROVIDERS.oauth_api.get_consumer(consumer_id)

        url = _update_url_scheme()
        req_headers = {'Requested-Project-Id': requested_project_id}
        req_headers.update(flask.request.headers)
        request_verifier = oauth1.RequestTokenEndpoint(
            request_validator=validator.OAuthValidator(),
            token_generator=oauth1.token_generator)
        h, b, s = request_verifier.create_request_token_response(
            url, http_method='POST', body=flask.request.args,
            headers=req_headers)
        if not b:
            msg = _('Invalid signature')
            raise exception.Unauthorized(message=msg)
        # show the details of the failure.
        oauth1.validate_oauth_params(b)
        request_token_duration = CONF.oauth1.request_token_duration
        token_ref = PROVIDERS.oauth_api.create_request_token(
            consumer_id,
            requested_project_id,
            request_token_duration,
            initiator=notifications.build_audit_initiator())

        result = ('oauth_token=%(key)s&oauth_token_secret=%(secret)s'
                  % {'key': token_ref['id'],
                     'secret': token_ref['request_secret']})

        if CONF.oauth1.request_token_duration > 0:
            expiry_bit = '&oauth_expires_at=%s' % token_ref['expires_at']
            result += expiry_bit

        resp = flask.make_response(result, http.client.CREATED)
        resp.headers['Content-Type'] = 'application/x-www-form-urlencoded'
        return resp


class AccessTokenResource(_OAuth1ResourceBase):
    @ks_flask.unenforced_api
    def post(self):
        oauth_headers = oauth1.get_oauth_headers(flask.request.headers)
        consumer_id = oauth_headers.get('oauth_consumer_key')
        request_token_id = oauth_headers.get('oauth_token')
        oauth_verifier = oauth_headers.get('oauth_verifier')

        if not consumer_id:
            raise exception.ValidationError(
                attribute='oauth_consumer_key', target='request')
        if not request_token_id:
            raise exception.ValidationError(
                attribute='oauth_token', target='request')
        if not oauth_verifier:
            raise exception.ValidationError(
                attribute='oauth_verifier', target='request')

        req_token = PROVIDERS.oauth_api.get_request_token(
            request_token_id)

        expires_at = req_token['expires_at']
        if expires_at:
            now = timeutils.utcnow()
            expires = timeutils.normalize_time(
                timeutils.parse_isotime(expires_at))
            if now > expires:
                raise exception.Unauthorized(_('Request token is expired'))

        url = _update_url_scheme()
        access_verifier = oauth1.AccessTokenEndpoint(
            request_validator=validator.OAuthValidator(),
            token_generator=oauth1.token_generator)
        try:
            h, b, s = access_verifier.create_access_token_response(
                url,
                http_method='POST',
                body=flask.request.args,
                headers=dict(flask.request.headers))
        except NotImplementedError:
            # Client key or request token validation failed, since keystone
            # does not yet support dummy client or dummy request token,
            # so we will raise unauthorized exception instead.
            try:
                PROVIDERS.oauth_api.get_consumer(consumer_id)
            except exception.NotFound:
                msg = _('Provided consumer does not exist.')
                LOG.warning('Provided consumer does not exist.')
                raise exception.Unauthorized(message=msg)
            if req_token['consumer_id'] != consumer_id:
                msg = ('Provided consumer key does not match stored consumer '
                       'key.')
                tr_msg = _('Provided consumer key does not match stored '
                           'consumer key.')
                LOG.warning(msg)
                raise exception.Unauthorized(message=tr_msg)
        # The response body is empty since either one of the following reasons
        if not b:
            if req_token['verifier'] != oauth_verifier:
                msg = 'Provided verifier does not match stored verifier'
                tr_msg = _('Provided verifier does not match stored verifier')
            else:
                msg = 'Invalid signature'
                tr_msg = _('Invalid signature')
            LOG.warning(msg)
            raise exception.Unauthorized(message=tr_msg)
        # show the details of the failure
        oauth1.validate_oauth_params(b)
        if not req_token.get('authorizing_user_id'):
            msg = _('Request Token does not have an authorizing user id.')
            LOG.warning('Request Token does not have an authorizing user id.')
            raise exception.Unauthorized(message=msg)

        access_token_duration = CONF.oauth1.access_token_duration
        token_ref = PROVIDERS.oauth_api.create_access_token(
            request_token_id,
            access_token_duration,
            initiator=notifications.build_audit_initiator())

        result = ('oauth_token=%(key)s&oauth_token_secret=%(secret)s'
                  % {'key': token_ref['id'],
                     'secret': token_ref['access_secret']})

        if CONF.oauth1.access_token_duration > 0:
            expiry_bit = '&oauth_expires_at=%s' % (token_ref['expires_at'])
            result += expiry_bit

        resp = flask.make_response(result, http.client.CREATED)
        resp.headers['Content-Type'] = 'application/x-www-form-urlencoded'
        return resp


class AuthorizeResource(_OAuth1ResourceBase):
    def put(self, request_token_id):
        ENFORCER.enforce_call(action='identity:authorize_request_token')
        roles = (flask.request.get_json(force=True, silent=True) or {}).get(
            'roles', [])
        validation.lazy_validate(schema.request_token_authorize, roles)
        ctx = flask.request.environ[context.REQUEST_CONTEXT_ENV]
        if ctx.is_delegated_auth:
            raise exception.Forbidden(
                _('Cannot authorize a request token with a token issued via '
                  'delegation.'))

        req_token = PROVIDERS.oauth_api.get_request_token(request_token_id)

        expires_at = req_token['expires_at']
        if expires_at:
            now = timeutils.utcnow()
            expires = timeutils.normalize_time(
                timeutils.parse_isotime(expires_at))
            if now > expires:
                raise exception.Unauthorized(_('Request token is expired'))

        authed_roles = _normalize_role_list(roles)

        # verify the authorizing user has the roles
        try:
            auth_context = flask.request.environ[
                authorization.AUTH_CONTEXT_ENV]
            user_token_ref = auth_context['token']
        except KeyError:
            LOG.warning("Couldn't find the auth context.")
            raise exception.Unauthorized()

        user_id = user_token_ref.user_id
        project_id = req_token['requested_project_id']
        user_roles = PROVIDERS.assignment_api.get_roles_for_user_and_project(
            user_id, project_id)
        cred_set = set(user_roles)

        if not cred_set.issuperset(authed_roles):
            msg = _('authorizing user does not have role required')
            raise exception.Unauthorized(message=msg)

        # create least of just the id's for the backend
        role_ids = list(authed_roles)

        # finally authorize the token
        authed_token = PROVIDERS.oauth_api.authorize_request_token(
            request_token_id, user_id, role_ids)

        to_return = {'token': {'oauth_verifier': authed_token['verifier']}}
        return to_return


class OSAuth1API(ks_flask.APIBase):
    _name = 'OS-OAUTH1'
    _import_name = __name__
    _api_url_prefix = '/OS-OAUTH1'
    resources = [ConsumerResource]
    resource_mapping = [
        ks_flask.construct_resource_map(
            resource=RequestTokenResource,
            url='/request_token',
            resource_kwargs={},
            rel='request_tokens',
            resource_relation_func=_build_resource_relation
        ),
        ks_flask.construct_resource_map(
            resource=AccessTokenResource,
            url='/access_token',
            rel='access_tokens',
            resource_kwargs={},
            resource_relation_func=_build_resource_relation
        ),
        ks_flask.construct_resource_map(
            resource=AuthorizeResource,
            url='/authorize/<string:request_token_id>',
            resource_kwargs={},
            rel='authorize_request_token',
            resource_relation_func=_build_resource_relation,
            path_vars={
                'request_token_id': _build_parameter_relation(
                    parameter_name='request_token_id')
            })]


APIs = (OSAuth1API,)



# Copyright 2014 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from keystone.common import sql
from keystone.identity.mapping_backends import base
from keystone.identity.mapping_backends import mapping as identity_mapping


class IDMapping(sql.ModelBase, sql.ModelDictMixin):
    __tablename__ = 'id_mapping'
    public_id = sql.Column(sql.String(64), primary_key=True)
    domain_id = sql.Column(sql.String(64), nullable=False)
    local_id = sql.Column(sql.String(255), nullable=False)
    # NOTE(henry-nash): Postgres requires a name to be defined for an Enum
    entity_type = sql.Column(
        sql.Enum(identity_mapping.EntityType.USER,
                 identity_mapping.EntityType.GROUP,
                 name='entity_type'),
        nullable=False)
    # Unique constraint to ensure you can't store more than one mapping to the
    # same underlying values
    __table_args__ = (
        sql.UniqueConstraint('domain_id', 'local_id', 'entity_type'),)


class Mapping(base.MappingDriverBase):

    def get_public_id(self, local_entity):
        # NOTE(henry-nash): Since the Public ID is regeneratable, rather
        # than search for the entry using the local entity values, we
        # could create the hash and do a PK lookup.  However this would only
        # work if we hashed all the entries, even those that already generate
        # UUIDs, like SQL.  Further, this would only work if the generation
        # algorithm was immutable (e.g. it had always been sha256).
        with sql.session_for_read() as session:
            query = session.query(IDMapping.public_id)
            query = query.filter_by(domain_id=local_entity['domain_id'])
            query = query.filter_by(local_id=local_entity['local_id'])
            query = query.filter_by(entity_type=local_entity['entity_type'])
            try:
                public_ref = query.one()
                public_id = public_ref.public_id
                return public_id
            except sql.NotFound:
                return None

    def get_domain_mapping_list(self, domain_id, entity_type=None):
        filters = {'domain_id': domain_id}
        if entity_type is not None:
            filters['entity_type'] = entity_type
        with sql.session_for_read() as session:
            return session.query(IDMapping).filter_by(**filters)

    def get_id_mapping(self, public_id):
        with sql.session_for_read() as session:
            mapping_ref = session.query(IDMapping).get(public_id)
            if mapping_ref:
                return mapping_ref.to_dict()

    def create_id_mapping(self, local_entity, public_id=None):
        entity = local_entity.copy()
        try:
            with sql.session_for_write() as session:
                if public_id is None:
                    public_id = self.id_generator_api.generate_public_ID(
                        entity)
                entity['public_id'] = public_id
                mapping_ref = IDMapping.from_dict(entity)
                session.add(mapping_ref)
        except sql.DBDuplicateEntry:
            # something else created the mapping already. We can use it.
            public_id = self.get_public_id(local_entity)
        return public_id

    def delete_id_mapping(self, public_id):
        with sql.session_for_write() as session:
            try:
                session.query(IDMapping).filter(
                    IDMapping.public_id == public_id).delete()
            except sql.NotFound:  # nosec
                # NOTE(morganfainberg): There is nothing to delete and nothing
                # to do.
                pass

    def purge_mappings(self, purge_filter):
        with sql.session_for_write() as session:
            query = session.query(IDMapping)
            if 'domain_id' in purge_filter:
                query = query.filter_by(domain_id=purge_filter['domain_id'])
            if 'public_id' in purge_filter:
                query = query.filter_by(public_id=purge_filter['public_id'])
            if 'local_id' in purge_filter:
                query = query.filter_by(local_id=purge_filter['local_id'])
            if 'entity_type' in purge_filter:
                query = query.filter_by(
                    entity_type=purge_filter['entity_type'])
            query.delete()



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import uuid

import http.client
from oslo_serialization import jsonutils

from keystone.common.policies import base as bp
from keystone.common import provider_api
import keystone.conf
from keystone.tests.common import auth as common_auth
from keystone.tests import unit
from keystone.tests.unit import base_classes
from keystone.tests.unit import ksfixtures
from keystone.tests.unit.ksfixtures import temporaryfile

CONF = keystone.conf.CONF
PROVIDERS = provider_api.ProviderAPIs


class _UserCredentialTests(object):
    """Test cases for anyone that has a valid user token."""

    def test_user_can_create_credentials_for_themselves(self):
        create = {
            'credential': {
                'blob': uuid.uuid4().hex,
                'user_id': self.user_id,
                'type': uuid.uuid4().hex
            }
        }
        with self.test_client() as c:
            c.post('/v3/credentials', json=create, headers=self.headers)

    def test_user_can_get_their_credentials(self):
        with self.test_client() as c:
            create = {
                'credential': {
                    'blob': uuid.uuid4().hex,
                    'type': uuid.uuid4().hex,
                    'user_id': self.user_id
                }
            }
            r = c.post('/v3/credentials', json=create, headers=self.headers)
            credential_id = r.json['credential']['id']

            path = '/v3/credentials/%s' % credential_id
            r = c.get(path, headers=self.headers)
            self.assertEqual(
                self.user_id, r.json['credential']['user_id']
            )

    def test_user_can_list_their_credentials(self):
        with self.test_client() as c:
            expected = []
            for _ in range(2):
                create = {
                    'credential': {
                        'blob': uuid.uuid4().hex,
                        'type': uuid.uuid4().hex,
                        'user_id': self.user_id
                    }
                }
                r = c.post(
                    '/v3/credentials', json=create, headers=self.headers
                )
                expected.append(r.json['credential'])

            r = c.get('/v3/credentials', headers=self.headers)
            for credential in expected:
                self.assertIn(credential, r.json['credentials'])

    def test_user_can_filter_their_credentials_by_type_and_user(self):
        with self.test_client() as c:
            credential_type = uuid.uuid4().hex
            create = {
                'credential': {
                    'blob': uuid.uuid4().hex,
                    'type': credential_type,
                    'user_id': self.user_id
                }
            }
            r = c.post(
                '/v3/credentials', json=create, headers=self.headers
            )
            expected_credential_id = r.json['credential']['id']

            create = {
                'credential': {
                    'blob': uuid.uuid4().hex,
                    'type': uuid.uuid4().hex,
                    'user_id': self.user_id
                }
            }
            r = c.post(
                '/v3/credentials', json=create, headers=self.headers
            )

            path = '/v3/credentials?type=%s' % credential_type
            r = c.get(path, headers=self.headers)
            self.assertEqual(
                expected_credential_id, r.json['credentials'][0]['id']
            )

            path = '/v3/credentials?user=%s' % self.user_id
            r = c.get(path, headers=self.headers)
            self.assertEqual(
                expected_credential_id, r.json['credentials'][0]['id']
            )

    def test_user_can_update_their_credential(self):
        with self.test_client() as c:
            create = {
                'credential': {
                    'blob': uuid.uuid4().hex,
                    'type': uuid.uuid4().hex,
                    'user_id': self.user_id
                }
            }

            r = c.post('/v3/credentials', json=create, headers=self.headers)
            credential_id = r.json['credential']['id']

            updated_blob = uuid.uuid4().hex
            update = {'credential': {'blob': updated_blob}}
            path = '/v3/credentials/%s' % credential_id
            r = c.patch(path, json=update, headers=self.headers)
            self.assertEqual(updated_blob, r.json['credential']['blob'])

    def test_user_can_delete_their_credentials(self):
        with self.test_client() as c:
            create = {
                'credential': {
                    'blob': uuid.uuid4().hex,
                    'type': uuid.uuid4().hex,
                    'user_id': self.user_id
                }
            }
            r = c.post('/v3/credentials', json=create, headers=self.headers)
            credential_id = r.json['credential']['id']

            path = '/v3/credentials/%s' % credential_id
            c.delete(path, headers=self.headers)


class _ProjectUsersTests(object):
    """Users who have project role authorization observe the same behavior."""

    def test_user_cannot_get_credentials_for_other_users(self):
        user = unit.new_user_ref(domain_id=CONF.identity.default_domain_id)
        user_password = user['password']
        user = PROVIDERS.identity_api.create_user(user)
        project = unit.new_project_ref(
            domain_id=CONF.identity.default_domain_id
        )
        project = PROVIDERS.resource_api.create_project(project['id'], project)
        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.member_role_id, user_id=user['id'],
            project_id=project['id']
        )
        user_auth = self.build_authentication_request(
            user_id=user['id'], password=user_password,
            project_id=project['id']
        )

        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=user_auth)
            token_id = r.headers['X-Subject-Token']
            headers = {'X-Auth-Token': token_id}

            create = {
                'credential': {
                    'blob': uuid.uuid4().hex,
                    'type': uuid.uuid4().hex,
                    'user_id': user['id']
                }
            }
            r = c.post('/v3/credentials', json=create, headers=headers)
            credential_id = r.json['credential']['id']

        with self.test_client() as c:
            path = '/v3/credentials/%s' % credential_id
            c.get(
                path, headers=self.headers,
                expected_status_code=http.client.FORBIDDEN
            )

    def test_user_cannot_get_non_existant_credential_forbidden(self):
        with self.test_client() as c:
            c.get(
                '/v3/credentials/%s' % uuid.uuid4().hex, headers=self.headers,
                expected_status_code=http.client.FORBIDDEN
            )

    def test_user_cannot_list_credentials_for_other_users(self):
        user = unit.new_user_ref(domain_id=CONF.identity.default_domain_id)
        user_password = user['password']
        user = PROVIDERS.identity_api.create_user(user)
        project = unit.new_project_ref(
            domain_id=CONF.identity.default_domain_id
        )
        project = PROVIDERS.resource_api.create_project(project['id'], project)
        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.member_role_id, user_id=user['id'],
            project_id=project['id']
        )
        user_auth = self.build_authentication_request(
            user_id=user['id'], password=user_password,
            project_id=project['id']
        )

        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=user_auth)
            token_id = r.headers['X-Subject-Token']
            headers = {'X-Auth-Token': token_id}

            create = {
                'credential': {
                    'blob': uuid.uuid4().hex,
                    'type': uuid.uuid4().hex,
                    'user_id': user['id']
                }
            }
            c.post('/v3/credentials', json=create, headers=headers)

        with self.test_client() as c:
            path = '/v3/credentials?user_id=%s' % user['id']
            r = c.get(path, headers=self.headers)
            self.assertEqual([], r.json['credentials'])

    def test_user_cannot_filter_credentials_by_type_for_others(self):
        user = unit.new_user_ref(domain_id=CONF.identity.default_domain_id)
        user_password = user['password']
        user = PROVIDERS.identity_api.create_user(user)
        project = unit.new_project_ref(
            domain_id=CONF.identity.default_domain_id
        )
        project = PROVIDERS.resource_api.create_project(project['id'], project)
        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.member_role_id, user_id=user['id'],
            project_id=project['id']
        )
        user_auth = self.build_authentication_request(
            user_id=user['id'], password=user_password,
            project_id=project['id']
        )

        credential_type = uuid.uuid4().hex
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=user_auth)
            token_id = r.headers['X-Subject-Token']
            headers = {'X-Auth-Token': token_id}

            create = {
                'credential': {
                    'blob': uuid.uuid4().hex,
                    'type': credential_type,
                    'user_id': user['id']
                }
            }
            c.post('/v3/credentials', json=create, headers=headers)

        with self.test_client() as c:
            path = '/v3/credentials?type=%s' % credential_type
            r = c.get(path, headers=self.headers)
            self.assertEqual(0, len(r.json['credentials']))

    def test_user_cannot_filter_credentials_by_user_for_others(self):
        user = unit.new_user_ref(domain_id=CONF.identity.default_domain_id)
        user_password = user['password']
        user = PROVIDERS.identity_api.create_user(user)
        project = unit.new_project_ref(
            domain_id=CONF.identity.default_domain_id
        )
        project = PROVIDERS.resource_api.create_project(project['id'], project)
        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.member_role_id, user_id=user['id'],
            project_id=project['id']
        )
        user_auth = self.build_authentication_request(
            user_id=user['id'], password=user_password,
            project_id=project['id']
        )

        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=user_auth)
            token_id = r.headers['X-Subject-Token']
            headers = {'X-Auth-Token': token_id}

            expected_cred_ids = []
            for _ in range(2):
                create = {
                    'credential': {
                        'blob': uuid.uuid4().hex,
                        'type': uuid.uuid4().hex,
                        'user_id': user['id']
                    }
                }
                r = c.post('/v3/credentials', json=create, headers=headers)
                expected_cred_ids.append(r.json['credential']['id'])

        with self.test_client() as c:
            path = '/v3/credentials?user_id=%s' % user['id']
            r = c.get(path, headers=self.headers)
            self.assertEqual([], r.json['credentials'])

    def test_user_cannot_update_credentials_for_others(self):
        user = unit.new_user_ref(domain_id=CONF.identity.default_domain_id)
        user_password = user['password']
        user = PROVIDERS.identity_api.create_user(user)
        project = unit.new_project_ref(
            domain_id=CONF.identity.default_domain_id
        )
        project = PROVIDERS.resource_api.create_project(project['id'], project)
        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.member_role_id, user_id=user['id'],
            project_id=project['id']
        )
        user_auth = self.build_authentication_request(
            user_id=user['id'], password=user_password,
            project_id=project['id']
        )

        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=user_auth)
            token_id = r.headers['X-Subject-Token']
            headers = {'X-Auth-Token': token_id}

            create = {
                'credential': {
                    'blob': uuid.uuid4().hex,
                    'type': uuid.uuid4().hex,
                    'user_id': user['id']
                }
            }
            r = c.post('/v3/credentials', json=create, headers=headers)
            credential_id = r.json['credential']['id']

        with self.test_client() as c:
            update = {'credential': {'blob': uuid.uuid4().hex}}
            path = '/v3/credentials/%s' % credential_id
            c.patch(
                path, json=update, headers=self.headers,
                expected_status_code=http.client.FORBIDDEN
            )

    def test_user_cannot_update_non_existant_credential_forbidden(self):
        with self.test_client() as c:
            update = {'credential': {'blob': uuid.uuid4().hex}}

            c.patch(
                '/v3/credentials/%s' % uuid.uuid4().hex, json=update,
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN
            )

    def test_user_cannot_create_credentials_for_other_users(self):
        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=CONF.identity.default_domain_id)
        )

        with self.test_client() as c:
            create = {
                'credential': {
                    'blob': uuid.uuid4().hex,
                    'type': uuid.uuid4().hex,
                    'user_id': user['id']
                }
            }
            c.post(
                '/v3/credentials', json=create, headers=self.headers,
                expected_status_code=http.client.FORBIDDEN
            )

    def test_user_cannot_delete_credentials_for_others(self):
        user = unit.new_user_ref(domain_id=CONF.identity.default_domain_id)
        user_password = user['password']
        user = PROVIDERS.identity_api.create_user(user)
        project = unit.new_project_ref(
            domain_id=CONF.identity.default_domain_id
        )
        project = PROVIDERS.resource_api.create_project(project['id'], project)
        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.member_role_id, user_id=user['id'],
            project_id=project['id']
        )
        user_auth = self.build_authentication_request(
            user_id=user['id'], password=user_password,
            project_id=project['id']
        )

        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=user_auth)
            token_id = r.headers['X-Subject-Token']
            headers = {'X-Auth-Token': token_id}

            create = {
                'credential': {
                    'blob': uuid.uuid4().hex,
                    'type': uuid.uuid4().hex,
                    'user_id': user['id']
                }
            }
            r = c.post('/v3/credentials', json=create, headers=headers)
            credential_id = r.json['credential']['id']

        with self.test_client() as c:
            path = '/v3/credentials/%s' % credential_id
            c.delete(
                path, headers=self.headers,
                expected_status_code=http.client.FORBIDDEN
            )

    def test_user_cannot_delete_non_existant_credential_forbidden(self):
        with self.test_client() as c:
            c.delete(
                '/v3/credentials/%s' % uuid.uuid4().hex, headers=self.headers,
                expected_status_code=http.client.FORBIDDEN
            )


class _SystemUserCredentialTests(object):
    """Tests that are common across all system users."""

    def test_user_can_list_credentials_for_other_users(self):
        user = unit.new_user_ref(domain_id=CONF.identity.default_domain_id)
        user_password = user['password']
        user = PROVIDERS.identity_api.create_user(user)
        project = unit.new_project_ref(
            domain_id=CONF.identity.default_domain_id
        )
        project = PROVIDERS.resource_api.create_project(project['id'], project)
        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.member_role_id, user_id=user['id'],
            project_id=project['id']
        )
        user_auth = self.build_authentication_request(
            user_id=user['id'], password=user_password,
            project_id=project['id']
        )

        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=user_auth)
            token_id = r.headers['X-Subject-Token']
            headers = {'X-Auth-Token': token_id}

            create = {
                'credential': {
                    'blob': uuid.uuid4().hex,
                    'type': uuid.uuid4().hex,
                    'user_id': user['id']
                }
            }
            r = c.post('/v3/credentials', json=create, headers=headers)
            credential_id = r.json['credential']['id']

        with self.test_client() as c:
            r = c.get('/v3/credentials', headers=self.headers)
            self.assertEqual(1, len(r.json['credentials']))
            self.assertEqual(credential_id, r.json['credentials'][0]['id'])
            self.assertEqual(user['id'], r.json['credentials'][0]['user_id'])

    def test_user_cannot_get_non_existant_credential_not_found(self):
        with self.test_client() as c:
            c.get(
                '/v3/credentials/%s' % uuid.uuid4().hex, headers=self.headers,
                expected_status_code=http.client.NOT_FOUND
            )

    def test_user_can_filter_credentials_by_type_for_others(self):
        user = unit.new_user_ref(domain_id=CONF.identity.default_domain_id)
        user_password = user['password']
        user = PROVIDERS.identity_api.create_user(user)
        project = unit.new_project_ref(
            domain_id=CONF.identity.default_domain_id
        )
        project = PROVIDERS.resource_api.create_project(project['id'], project)
        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.member_role_id, user_id=user['id'],
            project_id=project['id']
        )
        user_auth = self.build_authentication_request(
            user_id=user['id'], password=user_password,
            project_id=project['id']
        )

        credential_type = uuid.uuid4().hex
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=user_auth)
            token_id = r.headers['X-Subject-Token']
            headers = {'X-Auth-Token': token_id}

            create = {
                'credential': {
                    'blob': uuid.uuid4().hex,
                    'type': credential_type,
                    'user_id': user['id']
                }
            }
            r = c.post('/v3/credentials', json=create, headers=headers)
            credential_id = r.json['credential']['id']

            create = {
                'credential': {
                    'blob': uuid.uuid4().hex,
                    'type': uuid.uuid4().hex,
                    'user_id': user['id']
                }
            }
            c.post('/v3/credentials', json=create, headers=headers)

        with self.test_client() as c:
            path = '/v3/credentials?type=%s' % credential_type
            r = c.get(path, headers=self.headers)
            self.assertEqual(1, len(r.json['credentials']))
            self.assertEqual(credential_id, r.json['credentials'][0]['id'])
            self.assertEqual(user['id'], r.json['credentials'][0]['user_id'])

    def test_user_can_filter_credentials_by_user_for_others(self):
        user = unit.new_user_ref(domain_id=CONF.identity.default_domain_id)
        user_password = user['password']
        user = PROVIDERS.identity_api.create_user(user)
        project = unit.new_project_ref(
            domain_id=CONF.identity.default_domain_id
        )
        project = PROVIDERS.resource_api.create_project(project['id'], project)
        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.member_role_id, user_id=user['id'],
            project_id=project['id']
        )
        user_auth = self.build_authentication_request(
            user_id=user['id'], password=user_password,
            project_id=project['id']
        )

        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=user_auth)
            token_id = r.headers['X-Subject-Token']
            headers = {'X-Auth-Token': token_id}

            expected_cred_ids = []
            for _ in range(2):
                create = {
                    'credential': {
                        'blob': uuid.uuid4().hex,
                        'type': uuid.uuid4().hex,
                        'user_id': user['id']
                    }
                }
                r = c.post('/v3/credentials', json=create, headers=headers)
                expected_cred_ids.append(r.json['credential']['id'])

        with self.test_client() as c:
            path = '/v3/credentials?user_id=%s' % user['id']
            r = c.get(path, headers=self.headers)
            self.assertEqual(2, len(r.json['credentials']))
            for credential in r.json['credentials']:
                self.assertIn(credential['id'], expected_cred_ids)
                self.assertEqual(user['id'], credential['user_id'])


class SystemReaderTests(base_classes.TestCaseWithBootstrap,
                        common_auth.AuthTestMixin,
                        _UserCredentialTests,
                        _SystemUserCredentialTests):

    def setUp(self):
        super(SystemReaderTests, self).setUp()
        self.loadapp()
        self.useFixture(ksfixtures.Policy(self.config_fixture))
        self.config_fixture.config(group='oslo_policy', enforce_scope=True)

        system_reader = unit.new_user_ref(
            domain_id=CONF.identity.default_domain_id
        )
        self.user_id = PROVIDERS.identity_api.create_user(
            system_reader
        )['id']
        PROVIDERS.assignment_api.create_system_grant_for_user(
            self.user_id, self.bootstrapper.reader_role_id
        )

        auth = self.build_authentication_request(
            user_id=self.user_id, password=system_reader['password'],
            system=True
        )

        # Grab a token using the persona we're testing and prepare headers
        # for requests we'll be making in the tests.
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            self.token_id = r.headers['X-Subject-Token']
            self.headers = {'X-Auth-Token': self.token_id}

    def test_user_cannot_create_credentials_for_other_users(self):
        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=CONF.identity.default_domain_id)
        )

        with self.test_client() as c:
            create = {
                'credential': {
                    'blob': uuid.uuid4().hex,
                    'type': uuid.uuid4().hex,
                    'user_id': user['id']
                }
            }
            c.post(
                '/v3/credentials', json=create, headers=self.headers,
                expected_status_code=http.client.FORBIDDEN
            )

    def test_user_cannot_update_credentials_for_others(self):
        user = unit.new_user_ref(domain_id=CONF.identity.default_domain_id)
        user_password = user['password']
        user = PROVIDERS.identity_api.create_user(user)
        project = unit.new_project_ref(
            domain_id=CONF.identity.default_domain_id
        )
        project = PROVIDERS.resource_api.create_project(project['id'], project)
        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.member_role_id, user_id=user['id'],
            project_id=project['id']
        )
        user_auth = self.build_authentication_request(
            user_id=user['id'], password=user_password,
            project_id=project['id']
        )

        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=user_auth)
            token_id = r.headers['X-Subject-Token']
            headers = {'X-Auth-Token': token_id}

            create = {
                'credential': {
                    'blob': uuid.uuid4().hex,
                    'type': uuid.uuid4().hex,
                    'user_id': user['id']
                }
            }
            r = c.post('/v3/credentials', json=create, headers=headers)
            credential_id = r.json['credential']['id']

        with self.test_client() as c:
            update = {'credential': {'blob': uuid.uuid4().hex}}
            path = '/v3/credentials/%s' % credential_id
            c.patch(
                path, json=update, headers=self.headers,
                expected_status_code=http.client.FORBIDDEN
            )

    def test_user_cannot_update_non_existant_credential_forbidden(self):
        with self.test_client() as c:
            update = {'credential': {'blob': uuid.uuid4().hex}}

            c.patch(
                '/v3/credentials/%s' % uuid.uuid4().hex, json=update,
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN
            )

    def test_user_cannot_delete_credentials_for_others(self):
        user = unit.new_user_ref(domain_id=CONF.identity.default_domain_id)
        user_password = user['password']
        user = PROVIDERS.identity_api.create_user(user)
        project = unit.new_project_ref(
            domain_id=CONF.identity.default_domain_id
        )
        project = PROVIDERS.resource_api.create_project(project['id'], project)
        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.member_role_id, user_id=user['id'],
            project_id=project['id']
        )
        user_auth = self.build_authentication_request(
            user_id=user['id'], password=user_password,
            project_id=project['id']
        )

        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=user_auth)
            token_id = r.headers['X-Subject-Token']
            headers = {'X-Auth-Token': token_id}

            create = {
                'credential': {
                    'blob': uuid.uuid4().hex,
                    'type': uuid.uuid4().hex,
                    'user_id': user['id']
                }
            }
            r = c.post('/v3/credentials', json=create, headers=headers)
            credential_id = r.json['credential']['id']

        with self.test_client() as c:
            path = '/v3/credentials/%s' % credential_id
            c.delete(
                path, headers=self.headers,
                expected_status_code=http.client.FORBIDDEN
            )

    def test_user_cannot_delete_non_existant_credential_forbidden(self):
        with self.test_client() as c:
            c.delete(
                '/v3/credentials/%s' % uuid.uuid4().hex, headers=self.headers,
                expected_status_code=http.client.FORBIDDEN
            )


class SystemMemberTests(base_classes.TestCaseWithBootstrap,
                        common_auth.AuthTestMixin,
                        _UserCredentialTests,
                        _SystemUserCredentialTests):

    def setUp(self):
        super(SystemMemberTests, self).setUp()
        self.loadapp()
        self.useFixture(ksfixtures.Policy(self.config_fixture))
        self.config_fixture.config(group='oslo_policy', enforce_scope=True)

        system_member = unit.new_user_ref(
            domain_id=CONF.identity.default_domain_id
        )
        self.user_id = PROVIDERS.identity_api.create_user(
            system_member
        )['id']
        PROVIDERS.assignment_api.create_system_grant_for_user(
            self.user_id, self.bootstrapper.member_role_id
        )

        auth = self.build_authentication_request(
            user_id=self.user_id, password=system_member['password'],
            system=True
        )

        # Grab a token using the persona we're testing and prepare headers
        # for requests we'll be making in the tests.
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            self.token_id = r.headers['X-Subject-Token']
            self.headers = {'X-Auth-Token': self.token_id}

    def test_user_cannot_create_credentials_for_other_users(self):
        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=CONF.identity.default_domain_id)
        )

        with self.test_client() as c:
            create = {
                'credential': {
                    'blob': uuid.uuid4().hex,
                    'type': uuid.uuid4().hex,
                    'user_id': user['id']
                }
            }
            c.post(
                '/v3/credentials', json=create, headers=self.headers,
                expected_status_code=http.client.FORBIDDEN
            )

    def test_user_cannot_update_credentials_for_others(self):
        user = unit.new_user_ref(domain_id=CONF.identity.default_domain_id)
        user_password = user['password']
        user = PROVIDERS.identity_api.create_user(user)
        project = unit.new_project_ref(
            domain_id=CONF.identity.default_domain_id
        )
        project = PROVIDERS.resource_api.create_project(project['id'], project)
        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.member_role_id, user_id=user['id'],
            project_id=project['id']
        )
        user_auth = self.build_authentication_request(
            user_id=user['id'], password=user_password,
            project_id=project['id']
        )

        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=user_auth)
            token_id = r.headers['X-Subject-Token']
            headers = {'X-Auth-Token': token_id}

            create = {
                'credential': {
                    'blob': uuid.uuid4().hex,
                    'type': uuid.uuid4().hex,
                    'user_id': user['id']
                }
            }
            r = c.post('/v3/credentials', json=create, headers=headers)
            credential_id = r.json['credential']['id']

        with self.test_client() as c:
            update = {'credential': {'blob': uuid.uuid4().hex}}
            path = '/v3/credentials/%s' % credential_id
            c.patch(
                path, json=update, headers=self.headers,
                expected_status_code=http.client.FORBIDDEN
            )

    def test_user_cannot_update_non_existant_credential_forbidden(self):
        with self.test_client() as c:
            update = {'credential': {'blob': uuid.uuid4().hex}}

            c.patch(
                '/v3/credentials/%s' % uuid.uuid4().hex, json=update,
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN
            )

    def test_user_cannot_delete_credentials_for_others(self):
        user = unit.new_user_ref(domain_id=CONF.identity.default_domain_id)
        user_password = user['password']
        user = PROVIDERS.identity_api.create_user(user)
        project = unit.new_project_ref(
            domain_id=CONF.identity.default_domain_id
        )
        project = PROVIDERS.resource_api.create_project(project['id'], project)
        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.member_role_id, user_id=user['id'],
            project_id=project['id']
        )
        user_auth = self.build_authentication_request(
            user_id=user['id'], password=user_password,
            project_id=project['id']
        )

        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=user_auth)
            token_id = r.headers['X-Subject-Token']
            headers = {'X-Auth-Token': token_id}

            create = {
                'credential': {
                    'blob': uuid.uuid4().hex,
                    'type': uuid.uuid4().hex,
                    'user_id': user['id']
                }
            }
            r = c.post('/v3/credentials', json=create, headers=headers)
            credential_id = r.json['credential']['id']

        with self.test_client() as c:
            path = '/v3/credentials/%s' % credential_id
            c.delete(
                path, headers=self.headers,
                expected_status_code=http.client.FORBIDDEN
            )

    def test_user_cannot_delete_non_existant_credential_forbidden(self):
        with self.test_client() as c:
            c.delete(
                '/v3/credentials/%s' % uuid.uuid4().hex, headers=self.headers,
                expected_status_code=http.client.FORBIDDEN
            )


class SystemAdminTests(base_classes.TestCaseWithBootstrap,
                       common_auth.AuthTestMixin,
                       _UserCredentialTests,
                       _SystemUserCredentialTests):

    def setUp(self):
        super(SystemAdminTests, self).setUp()
        self.loadapp()
        self.useFixture(ksfixtures.Policy(self.config_fixture))
        self.config_fixture.config(group='oslo_policy', enforce_scope=True)

        # Reuse the system administrator account created during
        # ``keystone-manage bootstrap``
        self.user_id = self.bootstrapper.admin_user_id
        auth = self.build_authentication_request(
            user_id=self.user_id,
            password=self.bootstrapper.admin_password,
            system=True
        )

        # Grab a token using the persona we're testing and prepare headers
        # for requests we'll be making in the tests.
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            self.token_id = r.headers['X-Subject-Token']
            self.headers = {'X-Auth-Token': self.token_id}

    def test_user_can_create_credentials_for_other_users(self):
        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=CONF.identity.default_domain_id)
        )

        with self.test_client() as c:
            create = {
                'credential': {
                    'blob': uuid.uuid4().hex,
                    'type': uuid.uuid4().hex,
                    'user_id': user['id']
                }
            }
            c.post('/v3/credentials', json=create, headers=self.headers)

    def test_user_can_update_credentials_for_others(self):
        user = unit.new_user_ref(domain_id=CONF.identity.default_domain_id)
        user_password = user['password']
        user = PROVIDERS.identity_api.create_user(user)
        project = unit.new_project_ref(
            domain_id=CONF.identity.default_domain_id
        )
        project = PROVIDERS.resource_api.create_project(project['id'], project)
        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.member_role_id, user_id=user['id'],
            project_id=project['id']
        )
        user_auth = self.build_authentication_request(
            user_id=user['id'], password=user_password,
            project_id=project['id']
        )

        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=user_auth)
            token_id = r.headers['X-Subject-Token']
            headers = {'X-Auth-Token': token_id}

            create = {
                'credential': {
                    'blob': uuid.uuid4().hex,
                    'type': uuid.uuid4().hex,
                    'user_id': user['id']
                }
            }
            r = c.post('/v3/credentials', json=create, headers=headers)
            credential_id = r.json['credential']['id']

        with self.test_client() as c:
            path = '/v3/credentials/%s' % credential_id
            updated_blob = uuid.uuid4().hex
            update = {'credential': {'blob': updated_blob}}
            r = c.patch(path, json=update, headers=self.headers)
            self.assertEqual(updated_blob, r.json['credential']['blob'])
            self.assertEqual(user['id'], r.json['credential']['user_id'])

    def test_user_cannot_update_non_existant_credential_not_found(self):
        with self.test_client() as c:
            update = {'credential': {'blob': uuid.uuid4().hex}}

            c.patch(
                '/v3/credentials/%s' % uuid.uuid4().hex, json=update,
                headers=self.headers,
                expected_status_code=http.client.NOT_FOUND
            )

    def test_user_can_delete_credentials_for_others(self):
        user = unit.new_user_ref(domain_id=CONF.identity.default_domain_id)
        user_password = user['password']
        user = PROVIDERS.identity_api.create_user(user)
        project = unit.new_project_ref(
            domain_id=CONF.identity.default_domain_id
        )
        project = PROVIDERS.resource_api.create_project(project['id'], project)
        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.member_role_id, user_id=user['id'],
            project_id=project['id']
        )
        user_auth = self.build_authentication_request(
            user_id=user['id'], password=user_password,
            project_id=project['id']
        )

        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=user_auth)
            token_id = r.headers['X-Subject-Token']
            headers = {'X-Auth-Token': token_id}

            create = {
                'credential': {
                    'blob': uuid.uuid4().hex,
                    'type': uuid.uuid4().hex,
                    'user_id': user['id']
                }
            }
            r = c.post('/v3/credentials', json=create, headers=headers)
            credential_id = r.json['credential']['id']

        with self.test_client() as c:
            path = '/v3/credentials/%s' % credential_id
            c.delete(path, headers=self.headers)

    def test_user_cannot_delete_non_existant_credential_not_found(self):
        with self.test_client() as c:
            c.delete(
                '/v3/credentials/%s' % uuid.uuid4().hex, headers=self.headers,
                expected_status_code=http.client.NOT_FOUND
            )


class ProjectReaderTests(base_classes.TestCaseWithBootstrap,
                         common_auth.AuthTestMixin,
                         _UserCredentialTests,
                         _ProjectUsersTests):

    def setUp(self):
        super(ProjectReaderTests, self).setUp()
        self.loadapp()
        self.useFixture(ksfixtures.Policy(self.config_fixture))
        self.config_fixture.config(group='oslo_policy', enforce_scope=True)

        project_reader = unit.new_user_ref(
            domain_id=CONF.identity.default_domain_id
        )
        self.user_id = PROVIDERS.identity_api.create_user(
            project_reader
        )['id']
        project = unit.new_project_ref(
            domain_id=CONF.identity.default_domain_id
        )
        self.project_id = PROVIDERS.resource_api.create_project(
            project['id'], project
        )['id']
        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.reader_role_id, user_id=self.user_id,
            project_id=self.project_id
        )

        auth = self.build_authentication_request(
            user_id=self.user_id,
            password=project_reader['password'],
            project_id=self.project_id
        )

        # Grab a token using the persona we're testing and prepare headers
        # for requests we'll be making in the tests.
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            self.token_id = r.headers['X-Subject-Token']
            self.headers = {'X-Auth-Token': self.token_id}


class ProjectMemberTests(base_classes.TestCaseWithBootstrap,
                         common_auth.AuthTestMixin,
                         _UserCredentialTests,
                         _ProjectUsersTests):

    def setUp(self):
        super(ProjectMemberTests, self).setUp()
        self.loadapp()
        self.useFixture(ksfixtures.Policy(self.config_fixture))
        self.config_fixture.config(group='oslo_policy', enforce_scope=True)

        project_member = unit.new_user_ref(
            domain_id=CONF.identity.default_domain_id
        )
        self.user_id = PROVIDERS.identity_api.create_user(
            project_member
        )['id']
        project = unit.new_project_ref(
            domain_id=CONF.identity.default_domain_id
        )
        self.project_id = PROVIDERS.resource_api.create_project(
            project['id'], project
        )['id']
        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.member_role_id, user_id=self.user_id,
            project_id=self.project_id
        )

        auth = self.build_authentication_request(
            user_id=self.user_id,
            password=project_member['password'],
            project_id=self.project_id
        )

        # Grab a token using the persona we're testing and prepare headers
        # for requests we'll be making in the tests.
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            self.token_id = r.headers['X-Subject-Token']
            self.headers = {'X-Auth-Token': self.token_id}


class ProjectAdminTests(base_classes.TestCaseWithBootstrap,
                        common_auth.AuthTestMixin,
                        _UserCredentialTests,
                        _ProjectUsersTests):

    def setUp(self):
        super(ProjectAdminTests, self).setUp()
        self.loadapp()

        self.policy_file = self.useFixture(temporaryfile.SecureTempFile())
        self.policy_file_name = self.policy_file.file_name
        self.useFixture(
            ksfixtures.Policy(
                self.config_fixture, policy_file=self.policy_file_name
            )
        )
        self._override_policy()
        self.config_fixture.config(group='oslo_policy', enforce_scope=True)

        # Reuse the system administrator account created during
        # ``keystone-manage bootstrap``
        self.user_id = self.bootstrapper.admin_user_id
        auth = self.build_authentication_request(
            user_id=self.user_id,
            password=self.bootstrapper.admin_password,
            project_id=self.bootstrapper.project_id
        )

        # Grab a token using the persona we're testing and prepare headers
        # for requests we'll be making in the tests.
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            self.token_id = r.headers['X-Subject-Token']
            self.headers = {'X-Auth-Token': self.token_id}

    def _override_policy(self):
        # TODO(lbragstad): Remove this once the deprecated policies in
        # keystone.common.policies.credentials have been removed. This is only
        # here to make sure we test the new policies instead of the deprecated
        # ones. Oslo.policy will OR deprecated policies with new policies to
        # maintain compatibility and give operators a chance to update
        # permissions or update policies without breaking users. This will
        # cause these specific tests to fail since we're trying to correct this
        # broken behavior with better scope checking.
        with open(self.policy_file_name, 'w') as f:
            overridden_policies = {
                'identity:get_credential': bp.SYSTEM_READER_OR_CRED_OWNER,
                'identity:list_credentials': bp.SYSTEM_READER_OR_CRED_OWNER,
                'identity:create_credential': bp.SYSTEM_ADMIN_OR_CRED_OWNER,
                'identity:update_credential': bp.SYSTEM_ADMIN_OR_CRED_OWNER,
                'identity:delete_credential': bp.SYSTEM_ADMIN_OR_CRED_OWNER
            }
            f.write(jsonutils.dumps(overridden_policies))


class ProjectReaderTestsEnforceScopeFalse(base_classes.TestCaseWithBootstrap,
                                          common_auth.AuthTestMixin,
                                          _UserCredentialTests,
                                          _ProjectUsersTests):

    def setUp(self):
        super(ProjectReaderTestsEnforceScopeFalse, self).setUp()
        self.loadapp()
        self.useFixture(ksfixtures.Policy(self.config_fixture))
        self.config_fixture.config(group='oslo_policy', enforce_scope=False)

        project_reader = unit.new_user_ref(
            domain_id=CONF.identity.default_domain_id
        )
        self.user_id = PROVIDERS.identity_api.create_user(
            project_reader
        )['id']
        project = unit.new_project_ref(
            domain_id=CONF.identity.default_domain_id
        )
        self.project_id = PROVIDERS.resource_api.create_project(
            project['id'], project
        )['id']
        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.reader_role_id, user_id=self.user_id,
            project_id=self.project_id
        )

        auth = self.build_authentication_request(
            user_id=self.user_id,
            password=project_reader['password'],
            project_id=self.project_id
        )

        # Grab a token using the persona we're testing and prepare headers
        # for requests we'll be making in the tests.
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            self.token_id = r.headers['X-Subject-Token']
            self.headers = {'X-Auth-Token': self.token_id}


class ProjectMemberTestsEnforceScopeFalse(base_classes.TestCaseWithBootstrap,
                                          common_auth.AuthTestMixin,
                                          _UserCredentialTests,
                                          _ProjectUsersTests):

    def setUp(self):
        super(ProjectMemberTestsEnforceScopeFalse, self).setUp()
        self.loadapp()
        self.useFixture(ksfixtures.Policy(self.config_fixture))
        self.config_fixture.config(group='oslo_policy', enforce_scope=False)

        project_member = unit.new_user_ref(
            domain_id=CONF.identity.default_domain_id
        )
        self.user_id = PROVIDERS.identity_api.create_user(
            project_member
        )['id']
        project = unit.new_project_ref(
            domain_id=CONF.identity.default_domain_id
        )
        self.project_id = PROVIDERS.resource_api.create_project(
            project['id'], project
        )['id']
        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.member_role_id, user_id=self.user_id,
            project_id=self.project_id
        )

        auth = self.build_authentication_request(
            user_id=self.user_id,
            password=project_member['password'],
            project_id=self.project_id
        )

        # Grab a token using the persona we're testing and prepare headers
        # for requests we'll be making in the tests.
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            self.token_id = r.headers['X-Subject-Token']
            self.headers = {'X-Auth-Token': self.token_id}


class ProjectAdminTestsEnforceScopeFalse(base_classes.TestCaseWithBootstrap,
                                         common_auth.AuthTestMixin,
                                         _UserCredentialTests,
                                         _SystemUserCredentialTests):

    def setUp(self):
        super(ProjectAdminTestsEnforceScopeFalse, self).setUp()
        self.loadapp()
        self.useFixture(ksfixtures.Policy(self.config_fixture))
        self.config_fixture.config(group='oslo_policy', enforce_scope=False)

        # Reuse the system administrator account created during
        # ``keystone-manage bootstrap``
        self.user_id = self.bootstrapper.admin_user_id
        auth = self.build_authentication_request(
            user_id=self.user_id,
            password=self.bootstrapper.admin_password,
            project_id=self.bootstrapper.project_id
        )

        # Grab a token using the persona we're testing and prepare headers
        # for requests we'll be making in the tests.
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            self.token_id = r.headers['X-Subject-Token']
            self.headers = {'X-Auth-Token': self.token_id}



# Copyright 2013 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import uuid

from keystone.common import provider_api
import keystone.conf
from keystone import exception


PROVIDERS = provider_api.ProviderAPIs
CONF = keystone.conf.CONF


class FilterTests(object):

    # Provide support for checking if a batch of list items all
    # exist within a contiguous range in a total list
    def _match_with_list(self, this_batch, total_list,
                         batch_size=None,
                         list_start=None, list_end=None):
        if batch_size is None:
            batch_size = len(this_batch)
        if list_start is None:
            list_start = 0
        if list_end is None:
            list_end = len(total_list)
        for batch_item in range(0, batch_size):
            found = False
            for list_item in range(list_start, list_end):
                if this_batch[batch_item]['id'] == total_list[list_item]['id']:
                    found = True
            self.assertTrue(found)

    def _create_entity(self, entity_type):
        """Find the create_<entity_type> method.

        Searches through the [identity_api, resource_api, assignment_api]
        managers for a method called create_<entity_type> and returns the first
        one.

        """
        f = getattr(PROVIDERS.identity_api, 'create_%s' % entity_type, None)
        if f is None:
            f = getattr(
                PROVIDERS.resource_api, 'create_%s' % entity_type, None
            )
        if f is None:
            f = getattr(PROVIDERS.assignment_api, 'create_%s' % entity_type)
        return f

    def _delete_entity(self, entity_type):
        """Find the delete_<entity_type> method.

        Searches through the [identity_api, resource_api, assignment_api]
        managers for a method called delete_<entity_type> and returns the first
        one.

        """
        f = getattr(PROVIDERS.identity_api, 'delete_%s' % entity_type, None)
        if f is None:
            f = getattr(
                PROVIDERS.resource_api, 'delete_%s' % entity_type, None
            )
        if f is None:
            f = getattr(PROVIDERS.assignment_api, 'delete_%s' % entity_type)
        return f

    def _list_entities(self, entity_type):
        """Find the list_<entity_type> method.

        Searches through the [identity_api, resource_api, assignment_api]
        managers for a method called list_<entity_type> and returns the first
        one.

        """
        f = getattr(PROVIDERS.identity_api, 'list_%ss' % entity_type, None)
        if f is None:
            f = getattr(PROVIDERS.resource_api, 'list_%ss' % entity_type, None)
        if f is None:
            f = getattr(PROVIDERS.assignment_api, 'list_%ss' % entity_type)
        return f

    def _create_one_entity(self, entity_type, domain_id, name):
        new_entity = {'name': name,
                      'domain_id': domain_id}
        if entity_type in ['user', 'group']:
            # The manager layer creates the ID for users and groups
            new_entity = self._create_entity(entity_type)(new_entity)
        else:
            new_entity['id'] = uuid.uuid4().hex
            self._create_entity(entity_type)(new_entity['id'], new_entity)
        return new_entity

    def _create_test_data(self, entity_type, number, domain_id=None,
                          name_dict=None):
        """Create entity test data.

        :param entity_type: type of entity to create, e.g. 'user', group' etc.
        :param number: number of entities to create,
        :param domain_id: if not defined, all users will be created in the
                          default domain.
        :param name_dict: optional dict containing entity number and name pairs

        """
        entity_list = []
        if domain_id is None:
            domain_id = CONF.identity.default_domain_id
        name_dict = name_dict or {}
        for x in range(number):
            # If this index has a name defined in the name_dict, then use it
            name = name_dict.get(x, uuid.uuid4().hex)
            new_entity = self._create_one_entity(entity_type, domain_id, name)
            entity_list.append(new_entity)
        return entity_list

    def _delete_test_data(self, entity_type, entity_list):
        for entity in entity_list:
            try:
                self._delete_entity(entity_type)(entity['id'])
            except exception.Forbidden:
                # Note(knikolla): Some identity backends such as LDAP are
                # read only
                break



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

# All resource options are defined in this module. The individual resource
# implementations explicitly register the options that are desired directly
# in their individual registry. Each entry is imported from it's own
# module directly to allow for custom implementation details as needed.

from keystone.common.resource_options.options import immutable

__all__ = (
    'IMMUTABLE_OPT',
    'check_resource_immutable',
    'check_immutable_update',
    'check_immutable_delete',
)

# Immutable Option and helper functions
IMMUTABLE_OPT = immutable.IMMUTABLE_OPT
check_resource_immutable = immutable.check_resource_immutable
check_immutable_update = immutable.check_immutable_update
check_immutable_delete = immutable.check_immutable_delete
